"""Manage links beteween devices."""
import logging

from ...constants import AllLinkMode, LinkStatus, ResponseStatus
from ...handlers.start_all_linking import StartAllLinkingCommandHandler

TIMEOUT = 3
_LOGGER = logging.getLogger(__name__)


async def async_enter_linking_mode(is_controller: bool, group: int):
    """Put the Insteon Modem into linking mode."""
    link_cmd = StartAllLinkingCommandHandler()
    mode = AllLinkMode.CONTROLLER if is_controller else AllLinkMode.RESPONDER
    response = await link_cmd.async_send(mode=mode, group=group)
    _LOGGER.debug("Enter linking mode response: %s", str(response))
    return response


async def async_enter_unlinking_mode(group: int):
    """Put the Insteon Modem into unlinking mode."""
    link_cmd = StartAllLinkingCommandHandler()
    mode = AllLinkMode.DELETE
    response = await link_cmd.async_send(mode=mode, group=group)
    return response


async def async_link_devices(
    controller, responder, group: int = 0, data1=255, data2=28, data3=1
):
    """Link two devices."""

    if not hasattr(controller, "address") or not hasattr(responder, "address"):
        raise TypeError("controller and responder must be devices")
    if _add_link_to_device(
        device=controller,
        target=responder.address,
        group=group,
        is_controller=True,
        data1=responder.cat,
        data2=responder.subcat,
        data3=responder.firmware,
    ) and _add_link_to_device(
        device=responder,
        target=controller.address,
        group=group,
        is_controller=False,
        data1=data1,
        data2=data2,
        data3=data3,
    ):
        _, failed_1 = await controller.aldb.async_write()
        _, failed_2 = await responder.aldb.async_write()
        if failed_1 or failed_2:
            return ResponseStatus.FAILURE
        return ResponseStatus.SUCCESS


async def async_unlink_devices(controller, responder, group: int = 0):
    """Unlink two devices."""
    if not hasattr(controller, "address") or not hasattr(responder, "address"):
        raise TypeError("controller and responder must be devices")
    if _remove_link_from_device(
        device=controller, is_controller=True, group=group, target=responder.address
    ) and _remove_link_from_device(
        device=responder, is_controller=False, group=group, target=controller.address
    ):
        _, failed_1 = await controller.aldb.async_write()
        _, failed_2 = await responder.aldb.async_write()
        if failed_1 or failed_2:
            return ResponseStatus.FAILURE
        return ResponseStatus.SUCCESS
    return ResponseStatus.FAILURE


def _remove_link_from_device(device, is_controller, group, target):
    found_rec = None
    for addr in device.aldb:
        rec = device.aldb[addr]
        if (
            rec.target == target
            and rec.is_controller == is_controller
            and rec.group == group
        ):
            found_rec = rec
    if found_rec:
        device.aldb.remove(found_rec.mem_addr)
    return True


def find_broken_links(devices):
    """Find proken links."""
    broken_link_list = {}
    for addr in devices:
        device = devices[addr]
        for mem_addr in device.aldb:
            rec = device.aldb[mem_addr]
            if rec.is_in_use:
                status = _test_broken(addr, rec, devices)
                if status != LinkStatus.FOUND:
                    if not broken_link_list.get(addr):
                        broken_link_list[addr] = {}
                    broken_link_list[addr][mem_addr] = (rec, status)
    return broken_link_list


def _test_broken(address, rec, devices):
    """Test if a corresponding record exists in liked device."""
    device = devices.get(rec.target)
    if not device:
        return LinkStatus.MISSING_TARGET

    if not device.aldb.is_loaded:
        return LinkStatus.TARGET_DB_NOT_LOADED

    for mem_addr in device.aldb:
        t_rec = device.aldb[mem_addr]
        if (
            t_rec.target == address
            and t_rec.group == rec.group
            and t_rec.is_controller != rec.is_controller
        ):
            return LinkStatus.FOUND
    if rec.is_controller:
        return LinkStatus.MISSING_RESPONDER
    return LinkStatus.MISSING_CONTROLLER


def _add_link_to_device(device, is_controller, group, target, data1, data2, data3):
    """Add a link to a device."""
    try:
        device.aldb.add(
            group=group,
            target=target,
            controller=is_controller,
            data1=data1,
            data2=data2,
            data3=data3,
        )
    except NotImplementedError:
        return False
    return True



"""Representaton of an extended property of a device."""
from .address import Address
from .device_flag import DeviceFlagBase

RAMP_RATE = "ramp_rate"
X10_HOUSE = "x10_house"
X10_UNIT = "x10_unit"
LED_DIMMING = "led_dimming"
ON_LEVEL = "on_level"
AWAKE_INTERVAL = "awake_interval"
SLEEP_INTERVAL = "sleep_interval"
BROADCAST_NUMBER = "broadcast_number"
TRIGGER_GROUP_MASK = "trigger_group_mask"
LSB_OF_SLEEP_INTERVAL = "lsb_of_sleep_interval"
APP_RETRIES = "app_retries"
CONFIG = "config"
BATTERY_LEVEL = "battery_level"
DATABASE_DELTA = "database_delta"
SENSOR_STATUS = "sensor_status"
HEARBEAT_INTERVAL = "heartbeat_interval"
BATTERY_LOW_LEVEL = "battery_low_level"
LED_BRIGHTNESS = "led_brightness"
MOTION_TIMEOUT = "motion_timeout"
LIGHT_SENSITIVITY = "light_sensitivity"
HARDWARE_TIMEOUT = "hardware_timeout"
HARDWARE_LIGHT_SENSITIVITY = "hardware_light_sensitivity"
AMBIENT_LIGHT_INTENSITY = "ambient_light_intensity"
DELAY = "delay"
PRESCALER = "prescaler"
DURATION_HIGH = "duration_high"
DURATION_LOW = "duration_low"
ON_MASK = "on_mask"
OFF_MASK = "off_mask"
NON_TOGGLE_MASK = "non_toggle_mask"
NON_TOGGLE_ON_OFF_MASK = "non_toggle_on_off_mask"
BACKLIGHT = "backlight"
CHANGE_DELAY = "change_delay"
MASTER = "master"
HUMIDITY_HIGH = "humidity_high"
HUMIDITY_LOW = "humidity_low"
TEMP_OFFSET = "temp_offset"
TEMP_OFFSET_EXTERNAL = "temp_offset_external"
HUMIDITY_OFFSET = "humidity_offset"


class ExtendedProperty(DeviceFlagBase):
    """Representation of an extended property of a device."""

    def __init__(self, address, name, flag_type: type, is_reversed=False):
        """Init the ExtendedProperty class."""
        self._address = Address(address)
        topic = "{}.property.{}".format(self._address.id, name)
        super().__init__(topic, name, flag_type, is_reversed)



"""ALDB Read Manager."""
import asyncio
import logging

from async_generator import async_generator, yield_

from ..aldb.aldb_record import ALDBRecord
from ..handlers.from_device.receive_aldb_record import ReceiveALDBRecordHandler
from ..handlers.to_device.read_aldb import ReadALDBCommandHandler

RETRIES_ALL_MAX = 5
RETRIES_ONE_MAX = 20
RETRIES_WRITE_MAX = 5
TIMER = 5
TIMER_INCREMENT = 3
_LOGGER = logging.getLogger(__name__)
WRITE = 3
CANCEL = 0

IM_NOT_IN_DEVICE_ALDB = 0xFF
CHECKSUM_ERROR = 0xFD
ILLEGAL_VALUE_IN_COMMAND = 0xFB


def _is_multiple_records(mem_addr, num_recs):
    """Return true if we are searching for multiple records."""
    return (mem_addr == 0x00 and num_recs == 0) or num_recs > 1


class ALDBReadManager:
    """ALDB Read Manager."""

    def __init__(self, aldb, mem_addr: int = 0x00, num_recs: int = 0):
        """Init the ALDBReadManager class."""
        self._aldb = aldb
        self._mem_addr = mem_addr
        self._num_recs = num_recs

        self._records = asyncio.Queue()
        self._records_to_return = asyncio.Queue()
        self._retries_all = RETRIES_ALL_MAX
        self._retries_one = RETRIES_ONE_MAX
        self._last_mem_addr = 0x0000
        self._read_handler = ReadALDBCommandHandler(self._aldb.address)
        self._record_handler = ReceiveALDBRecordHandler(self._aldb.address)
        self._read_handler.subscribe(self._receive_direct_ack)
        self._record_handler.subscribe(self._receive_record)
        self._timer_lock = asyncio.Lock()

    @async_generator
    async def async_read(self, mem_addr: int = 0x00, num_recs: int = 0):
        """Start the reading process to enable iteration."""
        asyncio.ensure_future(
            self._async_read(mem_addr=self._mem_addr, num_recs=self._num_recs)
        )

        while True:
            record = await self._records_to_return.get()
            if record is None:
                return
            await yield_(record)

    async def _async_read(self, mem_addr: int = 0x00, num_recs: int = 0):
        """Perform the device read function."""
        # TODO check for success or failure
        _LOGGER.debug(
            "%s: Read memory %04x and %d records",
            str(self._aldb.address),
            mem_addr,
            num_recs,
        )

        if _is_multiple_records(mem_addr, num_recs):
            await self._read_all()
        else:
            await self._read_one(mem_addr)
        self._records_to_return.put_nowait(None)

    async def _read_one(self, mem_addr):
        """Read one record."""
        if self._aldb[mem_addr]:
            return True

        retries = RETRIES_ONE_MAX
        while retries:
            await self._read_handler.async_send(mem_addr=mem_addr, num_recs=1)

            timeout = TIMER + (RETRIES_ONE_MAX - retries) * TIMER_INCREMENT
            try:
                record = await self._get_next_record(timeout)
                if record is None:
                    return False
                if record.mem_addr == mem_addr:
                    return True
                await asyncio.sleep(0.1)
            except asyncio.TimeoutError:
                pass
            retries -= 1
        return False

    async def _read_all(self):
        """Read all records."""
        if self._aldb.is_loaded:
            return True

        retries = RETRIES_ALL_MAX
        while retries:
            await self._read_handler.async_send(mem_addr=0, num_recs=0)

            timeout = TIMER + (RETRIES_ALL_MAX - retries) * TIMER_INCREMENT
            try:
                while True:
                    record = await self._get_next_record(timeout)
                    if record is None:
                        return False
            except asyncio.TimeoutError:
                pass
            await asyncio.sleep(0.1)
            if self._aldb.is_loaded:
                return True
            retries -= 1

        # Read all records did not work so we try to read one at a time
        last_record = 0
        next_record = self._next_missing_record(last_record)
        while next_record is not None:
            if not await self._read_one(next_record):
                return False

            last_record = next_record
            next_record = self._next_missing_record(last_record)
            # If the next record equals the last record and we believe we successfully
            # read the last record, an error occured so we should just stop
            if next_record == last_record:
                return False
        return next_record is None

    async def _get_next_record(self, timeout):
        """Get the next record from the queue."""
        return await asyncio.wait_for(self._records.get(), timeout=timeout)

    def _receive_direct_ack(self, ack_response):
        """Receive the response from the direct ACK.

        The device has ackknowledged the request to read the ALDB.
        If the response is a negative response, stop the process.
        """
        if ack_response in [
            IM_NOT_IN_DEVICE_ALDB,
            CHECKSUM_ERROR,
            ILLEGAL_VALUE_IN_COMMAND,
        ]:
            _LOGGER.error(
                "%s: ALDB Load error: 0x%02x", str(self._aldb.address), ack_response
            )
            self._records.put_nowait(None)
            self._records_to_return.put_nowait(None)

    def _receive_record(
        self,
        memory,
        controller,
        group,
        target,
        data1,
        data2,
        data3,
        in_use,
        high_water_mark,
        bit5,
        bit4,
    ):
        """Receive an ALDB record."""
        record = ALDBRecord(
            memory=memory,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
        )
        self._records.put_nowait(record)
        self._records_to_return.put_nowait(record)

    def _next_missing_record(self, last_mem_addr):
        prev_addr = 0
        if not self._has_first_record():
            if last_mem_addr == 0x0000 and self._retries_one < RETRIES_ONE_MAX:
                return 0x0000
            return self._aldb.first_mem_addr

        for mem_addr in self._aldb:
            rec = self._aldb[mem_addr]
            if rec.is_high_water_mark:
                return None
            if prev_addr != 0:
                if not prev_addr - 8 == mem_addr:
                    return prev_addr - 8
            prev_addr = mem_addr
        next_addr = prev_addr - 8
        if (
            self._aldb.high_water_mark_mem_addr
            and next_addr <= self._aldb.high_water_mark_mem_addr
        ):
            return None
        return next_addr

    def _has_first_record(self):
        """Test if the first record is loaded."""
        for mem_addr in self._aldb:
            if mem_addr > 0x0FFF:
                return True
            if mem_addr in [self._aldb.first_mem_addr, 0x0FFF]:
                return True
        return False



"""Insteon device address class."""
import logging

from .constants import HC_LOOKUP
from .utils import (
    byte_to_housecode,
    byte_to_unitcode,
    housecode_to_byte,
    unitcode_to_byte,
)

_LOGGER = logging.getLogger(__name__)


def create(housecode: str, unitcode: int):
    """Create an X10 device address."""
    if housecode.lower() in HC_LOOKUP.keys():
        byte_housecode = housecode_to_byte(housecode)
    else:
        if isinstance(housecode, str):
            str_error = "X10 house code invalid: {}".format(housecode)
        else:
            str_error = "X10 house code is not a string"
            raise ValueError(str_error)

    # 20, 21 and 22 for All Units Off, All Lights On and All Lights Off
    # 'fake' units
    if unitcode in range(1, 17) or unitcode in range(20, 23):
        byte_unitcode = unitcode_to_byte(unitcode)
    else:
        if isinstance(unitcode, int):
            str_error = "X10 unit code error: {}".format(unitcode)
        else:
            str_error = "X10 unit code is not an integer 1 - 16"
            raise ValueError(str_error)

    addr = X10Address(bytearray([byte_housecode, byte_unitcode]))
    return addr


class X10Address:
    """Datatype definition for an X10 device address."""

    def __init__(self, housecode_unitcode: bytearray):
        """Create an X10 device address."""
        if isinstance(housecode_unitcode, X10Address):
            housecode_unitcode = bytes(housecode_unitcode)
        if len(housecode_unitcode) != 2:
            raise ValueError("housecode_unitcode must be 2 bytes")
        self._housecode_byte = housecode_unitcode[0]
        self._unitcode_byte = housecode_unitcode[1]
        valid = self._check_housecode_unitcode()
        if not valid:
            raise ValueError("Invalid housecode or unitcode byte")

    def __repr__(self):
        """Return the string representation of an X10 device address."""
        return "x10{}{}".format(
            byte_to_housecode(self._housecode_byte),
            byte_to_unitcode(self._unitcode_byte),
        )

    def __str__(self):
        """Return the string representation of an X10 device address."""
        str_rep = f"X10.{self.housecode}.{self.unitcode:02d}"
        return str(str_rep)

    def __bytes__(self):
        """Return the byte representation of an X10 address."""
        return bytes(bytearray([self._housecode_byte, self._unitcode_byte]))

    def __eq__(self, other):
        """Test if two X10 addresses are equal."""
        if isinstance(other, X10Address):
            return bytes(self) == bytes(other)
        return False

    def __getitem__(self, byte):
        """Return the housecode or unitcode bytes."""
        if byte == 0:
            return self.housecode_byte
        if byte == 1:
            return self.unitcode_byte
        err = "Item index must be 0 or 1: {}".format(byte)
        raise ValueError(err)

    def __hash__(self):
        """Return a hash of the address."""
        return hash(str)

    def _check_housecode_unitcode(self):
        """Check to confirm housecode and unitcode are valid."""
        housecode = byte_to_housecode(self._housecode_byte)
        unitcode = byte_to_unitcode(self._unitcode_byte)
        if housecode and unitcode:
            return True
        return False

    @property
    def id(self):
        """Return the ID of the X10 address."""
        return repr(self)

    @property
    def housecode_byte(self):
        """Emit the X10 house code byte value."""
        return self._housecode_byte

    @property
    def unitcode_byte(self):
        """Emit the X10 unit code byte value."""
        return self._unitcode_byte

    @property
    def housecode(self):
        """Emit the X10 house code."""
        return byte_to_housecode(self._housecode_byte).upper()

    @property
    def unitcode(self):
        """Emit the X10 unit code."""
        return byte_to_unitcode(self._unitcode_byte)



"""ALDB Records."""

from ..address import Address


class ALDBRecord:
    """Represents an ALDB record."""

    def __init__(
        self,
        memory: int,
        controller: bool,
        group: int,
        target: Address,
        data1: int,
        data2: int,
        data3: int,
        in_use: bool = True,
        high_water_mark: bool = False,
        bit5: bool = False,
        bit4: bool = False,
    ):
        """Initialze the ALDBRecord class."""
        self._memory_location = memory
        self._target = Address(target)
        self._group = group
        self._data1 = data1
        self._data2 = data2
        self._data3 = data3
        self._controller = controller
        self._in_use = in_use
        self._high_water_mark = high_water_mark
        self._bit5 = bit5
        self._bit4 = bit4

    def __str__(self):
        """Return the string representation of an ALDB record."""
        if self._controller:
            mode = "C"
        else:
            mode = "R"
        rec = {
            "memory": f"0x{self._memory_location:04x}",
            "inuse": self._in_use,
            "mode": mode,
            "bit5": str(self._bit5),
            "bit4": str(self._bit4),
            "highwater": str(self._high_water_mark),
            "group": f"0x{self.group:02x}",
            "target": str(self.target),
            "data1": f"0x{self.data1:02x}",
            "data2": f"0x{self.data2:02x}",
            "data3": f"0x{self.data3:02x}",
        }
        return str(rec)

    def __repr__(self):
        """Return a representation of the record."""
        rec = {
            "memory": f"0x{self._memory_location:04x}",
            "control_flags": f"0x{self.control_flags:02x}",
            "group": f"0x{self.group:02x}",
            "target": str(self.target),
            "data1": f"0x{self.data1:02x}",
            "data2": f"0x{self.data2:02x}",
            "data3": f"0x{self.data3:02x}",
        }
        return str(rec)

    def __dict___(self):
        """Return a dictionary object of the ALDB Record."""

        return {
            "d1": 0x00,
            "d2": 0x00,
            "d3": self.memhi,
            "d4": self.memlo,
            "d5": 0x00,
            "d6": self.control_flags,
            "d7": self.group,
            "d8": self.target.low,
            "d9": self.target.middle,
            "d10": self.target.high,
            "d11": self.data1,
            "d12": self.data2,
            "d13": self.data3,
            "d14": 0x00,
        }

    @property
    def mem_addr(self):
        """Return the memory address of the database record."""
        return self._memory_location

    @mem_addr.setter
    def mem_addr(self, value):
        """Set the memory address of the record."""
        try:
            mem = int(value)
        except ValueError:
            raise ValueError("Memory address must be an integer.")
        else:
            self._memory_location = mem

    @property
    def memhi(self):
        """Return the memory address MSB."""
        return self._memory_location >> 8

    @property
    def memlo(self):
        """Return the memory address LSB."""
        return self._memory_location & 0xFF

    @property
    def target(self):
        """Return the address of the device the record points to."""
        return self._target

    @property
    def group(self):
        """Return the group the record responds to."""
        return self._group

    @property
    def data1(self):
        """Return the data1 field of the ALDB record."""
        return self._data1

    @property
    def data2(self):
        """Return the data2 field of the ALDB record."""
        return self._data2

    @property
    def data3(self):
        """Return the data3 field of the ALDB record."""
        return self._data3

    @property
    def is_controller(self):
        """Return if the record is a controller record."""
        return self._controller

    @property
    def is_responder(self):
        """Return if the record is a responder."""
        return not self._controller

    @property
    def is_in_use(self):
        """Return if the record is in use."""
        return self._in_use

    @property
    def is_high_water_mark(self):
        """Return if this is the high water mark record."""
        return self._high_water_mark

    @property
    def is_bit5_set(self):
        """Return if control flag bit 5 is set."""
        return self._bit5

    @property
    def is_bit4_set(self):
        """Return if control flag bit 4 is set."""
        return self._bit4

    @property
    def control_flags(self):
        """Return the control flag byte."""
        return (
            int(self._in_use) << 7
            | int(self._controller) << 6
            | int(self._bit5) << 5
            | int(self._bit4) << 4
            | int(not self._high_water_mark) << 1
        )



"""Get and Set extended properties for a device."""
import asyncio
import logging
from collections import namedtuple

from ..address import Address
from ..constants import ResponseStatus
from ..extended_property import ExtendedProperty
from ..handlers.from_device.ext_get_response import ExtendedGetResponseHandler
from ..handlers.to_device.extended_get import ExtendedGetCommand
from ..handlers.to_device.extended_set import ExtendedSetCommand
from ..utils import multiple_status

_LOGGER = logging.getLogger(__name__)
TIMEOUT = 2
RETRIES = 20

PropertyInfo = namedtuple("PropertyInfo", "name group data_field bit set_cmd")


class GetSetExtendedPropertyManager:
    """Get and Set extended properties for a device."""

    def __init__(self, address: Address):
        """Init the GetSetExtendedPropertyManager class."""
        self._address = Address(address)
        self._properties = {}
        self._get_command = ExtendedGetCommand(address=self._address)
        self._get_response = ExtendedGetResponseHandler(address=self._address)
        self._get_response.subscribe(self._update_all_fields)
        self._groups = {}
        self._flags = {}
        self._response_queue = asyncio.Queue()
        self._get_cmd_lock = asyncio.Lock()

    @property
    def flag_info(self):
        """Return the flag information."""
        return self._flags

    def create(self, name, group, data_field, bit, set_cmd, is_revsersed=False):
        """Subscribe a device property to Get and Set values.

        data is stored in self._groups[group][data_field]<[bit]>
        """
        prop_type = bool if bit is not None else int
        flag_info = PropertyInfo(name, group, data_field, bit, set_cmd)
        flags = self._groups.get(group, {})
        field = flags.get(data_field, {})
        if bit is not None:
            field[bit] = flag_info
        else:
            field = flag_info
        flags[data_field] = field
        self._groups[group] = flags
        self._flags[name] = flag_info
        self._properties[name] = ExtendedProperty(
            self._address, name, prop_type, is_revsersed
        )
        return self._properties[name]

    async def async_read(self, group=None):
        """Get the properties for a group."""
        if self._get_cmd_lock.locked():
            self._get_cmd_lock.release()

        while not self._response_queue.empty():
            self._response_queue.get_nowait()

        await self._get_cmd_lock.acquire()
        if group is None:
            results = []
            for curr_group in self._groups:
                result = await self._async_read(group=curr_group)
                results.append(result)
                await asyncio.sleep(2)
            self._get_cmd_lock.release()
            return multiple_status(*results)
        result = await self._async_read(group=group)
        self._get_cmd_lock.release()
        return result

    async def _async_read(self, group):
        retry = 0
        result = ResponseStatus.UNSENT
        while retry < RETRIES and result != ResponseStatus.SUCCESS:
            await self._get_command.async_send(group=group)
            result = await self._wait_for_get()
            retry += 1
        return result

    async def async_write(self):
        """Set the properties for a group."""
        results = []
        for name in self._properties:
            if self._properties[name].is_dirty:
                group = self._flags[name].group
                data_field = self._flags[name].data_field
                result = await self._write_flag(name, group, data_field)
                results.append(result)
        return multiple_status(*results)

    async def _wait_for_get(self):
        """Wait for the get response message."""
        try:
            return await asyncio.wait_for(self._response_queue.get(), TIMEOUT)
        except asyncio.TimeoutError:
            return ResponseStatus.FAILURE

    async def _write_flag(self, name, group, field):
        """Write a flag to a device."""
        flag = self._properties[name]
        if isinstance(self._groups[group][field], PropertyInfo):
            data = flag.new_value
            set_cmd = self._groups[group][field].set_cmd
        else:
            set_cmd, data = self._calc_flag_value(self._groups[group][field])
        if set_cmd is not None:
            cmd = ExtendedSetCommand(self._address, data1=group, data2=set_cmd)
            result = await cmd.async_send(data3=data)
            if result == ResponseStatus.SUCCESS:
                self._update_one_field(group, field, data)
            return result
        return ResponseStatus.SUCCESS

    def _update_all_fields(self, group, data):
        """Update each flag."""
        if not self._get_cmd_lock.locked():
            return
        if self._groups.get(group) is None and group == 1:
            group = 0
        if self._groups.get(group) is None:
            return
        for field in self._groups[group]:
            value = data.get("data{}".format(field))
            self._update_one_field(group=group, field=field, value=value)
        self._response_queue.put_nowait(ResponseStatus.SUCCESS)

    def _update_one_field(self, group, field, value):
        """Update one field including bit based fields."""
        if isinstance(self._groups[group][field], PropertyInfo):
            flag_info = self._groups[group][field]
            flag = self._properties[flag_info.name]
            flag.load(value=value)
        else:
            for bit in self._groups[group][field]:
                flag_info = self._groups[group][field][bit]
                bit_value = bool(value & 1 << flag_info.bit)
                flag = self._properties[flag_info.name]
                flag.load(value=bit_value)
        self._response_queue.put_nowait(ResponseStatus.SUCCESS)

    def _calc_flag_value(self, field):
        data = 0x00
        set_cmd = None
        for bit in field:
            flag_info = field[bit]
            flag = self._properties[flag_info.name]
            set_value = not flag.new_value if flag.is_reversed else flag.new_value
            if set_value:
                data = data | 1 << bit
            set_cmd = flag_info.set_cmd
        return set_cmd, data



"""Fan On Level state."""
from .group_base import GroupBase
from ..address import Address
from ..constants import FanSpeed, FanSpeedRange


class FanOnLevel(GroupBase):
    """On / Off state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: FanSpeed = None
    ):
        """Init the FanOnLevel class."""
        super().__init__(name, address, group, default, value_type=FanSpeed)

    # pylint: disable=arguments-differ
    def set_value(self, on_level):
        """Set the value of the state from the handlers."""
        if on_level in FanSpeedRange.OFF:
            fan_speed = FanSpeed.OFF
        elif on_level in FanSpeedRange.LOW:
            fan_speed = FanSpeed.LOW
        elif on_level in FanSpeedRange.MEDIUM:
            fan_speed = FanSpeed.MEDIUM
        else:
            fan_speed = FanSpeed.HIGH
        self.value = fan_speed



"""Tools to run device commands."""

import inspect
from types import MethodType

from .tools_base import ToolsBase
from .. import devices
from ..managers.scene_manager import (
    async_trigger_scene_off,
    async_trigger_scene_on,
)


def _convert_val(val):
    if '"' not in val and "'" not in val:
        try:
            return int(val)
        except ValueError:
            try:
                return float(val)
            except ValueError:
                val = val.strip('"').strip("'")
    else:
        val = val.strip('"').strip("'")
    if val == "True":
        return True
    if val == "False":
        return False
    return val


def _parse_cmd(*args):
    try:
        address = args[0]
        if not devices[address]:
            raise IndexError
    except IndexError:
        return None, None, {}

    try:
        cmd = args[1]
        if not hasattr(devices[address], cmd):
            raise AttributeError
    except IndexError:
        return address, None, {}

    kwargs = {}
    for arg in args[2:]:
        kwarg, val = arg.split("=")
        val = _convert_val(val)
        kwargs[kwarg] = val

    return address, cmd, kwargs


class CmdTools(ToolsBase):
    """Tools to run device commands."""

    async def do_on(self, *args, **kwargs):
        """Test turning a device on.

        Usage:
            on <address>

        This method calls `async_on` with default values.
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return

        device = devices[addresses[0]]
        if not hasattr(device, "async_on"):
            self._log_stdout("Device cannot be turned on.")

        await device.async_on()

    async def do_off(self, *args, **kwargs):
        """Test turning a device off.

        Usage:
            on <address>

        This method calls `async_off` with default values.
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return

        device = devices[addresses[0]]
        if not hasattr(device, "async_off"):
            self._log_stdout("Device cannot be turned off.")

        await device.async_off()

    async def do_scene_on(self, *args, **kwargs):
        """Test a scene.

        Usage:
            scene_on scene
        """

        args = args[0].split()
        try:
            scene = int(args[0])
        except (IndexError, ValueError):
            scene = None

        if not scene:
            scene = await self._get_int(
                "Scene number or blank to cancel", values=range(25, 256),
            )

        if not scene:
            return

        await async_trigger_scene_on(scene)

    async def do_scene_off(self, *args, **kwargs):
        """Test a scene.

        Usage:
            scene_off scene
        """

        args = args[0].split()
        try:
            scene = int(args[0])
        except (IndexError, ValueError):
            scene = None

        if not scene:
            scene = await self._get_int(
                "Scene number or blank to cancel", values=range(25, 256),
            )

        if not scene:
            return

        await async_trigger_scene_off(scene)

    async def do_cmd(self, *args, **kwargs):
        """Run a general device command.

        Usage:
            cmd address command [arg1=value1 arg2=value2 ...]

        Examples:
            cmd 1a2b3c async_on on_level=100
            cmd 1a2b3c async_off

        No spaces between argument, '=' and value are allowed.
        If an enumerator is required (i.e. ThermostatMode) use the integer represenation of the enum.

        Example:
            cmd 1a2b3c async_set_mode mode=3

        where ThermostatMode.AUTO equals 3.

        """

        args = args[0].split()
        try:
            address, cmd, kwargs = _parse_cmd(*args)
        except AttributeError:
            self._log_stdout(f"Device command not found.")
            return

        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=False
        )
        if not addresses:
            return

        device = devices[addresses[0]]
        if not device:
            self._log_stdout("Device not found.")
            return

        if cmd is None:
            cmd = await self._get_char("Device command")
            if not hasattr(device, cmd):
                self._log_stdout("Device command not found.")

            if not kwargs:
                while True:
                    arg = self._get_char(
                        "Enter argument (eg. on_level=255). Press enter for last argument"
                    )
                    if arg == "":
                        break
                    kwarg, val = arg.split("=")
                    val = _convert_val(val)
                    kwargs[kwarg] = val

        self._log_stdout(f"cmd {address} {cmd} {kwargs}")

        func = getattr(device, cmd)
        if inspect.iscoroutinefunction(func) or inspect.isawaitable(func):
            result = await func(**kwargs)
        else:
            result = func(**kwargs)
        self._log_stdout(f"Result: {str(result)}")

    def do_help(self, arg):
        """List available commands with "help" or detailed help with "help cmd".

        You can also get a list of device commands to run with the "cmd" method.
        Example 1:
            help cmd 1a2b3c

        This will show the available commands that can be run against device 1A.2B.3C.

        Example 2:
            help cmd 1a2b3c async_off

        This will show the parameters available to the async_off method.
        """
        arg_set = arg.split()
        try:
            cmd = arg_set[0]
            if cmd != "cmd":
                return super().do_help(arg)

        except IndexError:
            return super().do_help(arg)

        try:
            address = arg_set[1]
        except IndexError:
            return super().do_help(arg)

        device = devices[address]
        if device:
            try:
                method = arg_set[2]
                if not hasattr(device, method):
                    self._log_stdout("Method not found for device.")
                    return
                func = getattr(device, method)
                arguments = str(inspect.signature(func)).strip("(").strip(")")
                self._log_stdout("")
                self._log_stdout(
                    f"Device {str(device.address)} {method} method arguments: {arguments}"
                )
            except IndexError:
                all_methods = dir(device)
                methods = []
                for method in all_methods:
                    if method[0] != "_":
                        func = getattr(device, method)
                        if isinstance(func, MethodType):
                            methods.append(method)
                self._log_stdout("")
                header = f"Available commands for device {str(device.address)}:"
                self.print_topics(header, methods, 15, 80)



"""Base class to handle Broadcast messages from devices."""
from ...address import Address
from ...constants import MessageFlagType
from ..inbound_base import InboundHandlerBase


class BroadcastCommandHandlerBase(InboundHandlerBase):
    """Base class to handle inbound Broadcast messages."""

    def __init__(self, topic, address, group=None):
        """Init the BroadcastHandlerBase class."""
        self._address = Address(address)
        self._group = group
        super().__init__(
            topic=topic,
            address=self._address,
            group=self._group,
            message_type=MessageFlagType.ALL_LINK_BROADCAST,
        )



"""Serial protocol to perform async I/O with the Powerline Modem (PLM)."""

import asyncio
import logging
from enum import Enum

from .. import pub
from ..utils import log_error, publish_topic, subscribe_topic
from .command_to_msg import register_command_handlers
from .messages.inbound import create
from .messages.outbound import register_outbound_handlers
from .msg_to_topic import convert_to_topic

_LOGGER = logging.getLogger(__name__)
_LOGGER_MSG = logging.getLogger("pyinsteon.messages")
WRITE_WAIT = 1.5  # Time to wait between writes to transport


def _is_nak(msg):
    """Test if a message is a NAK from the modem."""
    if hasattr(msg, "ack") and msg.ack.value == 0x15:
        return True
    return False


def _has_listeners(topic):
    """Test if a topic has listeners.

    Only used if the msg is a NAK. If no NAK specific listeners
    then resend the message. Otherwise it is assumed the NAK
    specific listner is resending if necessary.
    """
    topic_manager = pub.getDefaultTopicMgr()
    pub_topic = topic_manager.getTopic(name=topic, okIfNone=True)
    _LOGGER.debug("MSG Topic: %s", pub_topic)
    if pub_topic and pub_topic.getListeners():
        _LOGGER.debug("Has listeners so do not resend.")
        return True
    _LOGGER.debug("No listeners so resend")
    return False


class TransportStatus(Enum):
    """Status of the transport."""

    CLOSED = 0
    LOST = 1
    PAUSED = 2
    OPEN = 3


class Protocol(asyncio.Protocol):
    """Serial protocol to perform async I/O with the PLM."""

    def __init__(self, connect_method, *args, **kwargs):
        """Init the SerialProtocol class."""
        super().__init__(*args, **kwargs)
        self._transport = None
        self._message_queue = asyncio.PriorityQueue()
        self._buffer = bytearray()
        self._should_reconnect = True
        self._connect_method = connect_method
        register_outbound_handlers()
        register_command_handlers()
        subscribe_topic(self._write, "send_message")

    @property
    def connected(self) -> bool:
        """Return true if the transport is connected."""
        status = not self._transport.is_closing() if self._transport else False
        return status

    @property
    def message_queue(self):
        """Return the queue of messages to write to the transport."""
        return self._message_queue

    def connection_made(self, transport):
        """Run when a connection to the transport has been made."""
        self._transport = transport
        self._start_writer()
        publish_topic("connection.made")

    def data_received(self, data):
        """Receive data from the serial transport."""
        self._buffer.extend(data)
        while True:
            last_buffer = self._buffer
            msg, self._buffer = create(self._buffer)
            if msg:
                self._publish_message(msg)

            if last_buffer == self._buffer or not self._buffer:
                break

    def connection_lost(self, exc):
        """Notify listeners that the serial connection is lost."""
        _LOGGER.debug("Connection lost called")
        _LOGGER.debug("Should reconnect: %s", self._should_reconnect)
        self._stop_writer()
        if self._should_reconnect:
            asyncio.ensure_future(self.async_connect())

    async def async_connect(self, retry=True):
        """Connect to the transport asyncrously."""
        wait_time = 0.1
        while not self.connected:
            self._transport = await self._connect_method(protocol=self)
            if self._transport is None and not retry:
                raise ConnectionError("Modem did not respond to connection request")
            if not self.connected and retry:
                await asyncio.sleep(wait_time)
                wait_time = min(300, 1.5 * wait_time)
            else:
                return

    def pause_writing(self):
        """Pause writing to the transport."""
        self._stop_writer()

    def resume_writing(self):
        """Resume writing to the transport."""
        self._start_writer()

    def close(self):
        """Close the serial transport."""
        self._unsubscribe()
        self._should_reconnect = False
        self._stop_writer()
        if self._transport:
            self._transport.close()

    def _stop_writer(self):
        self._message_queue.put_nowait((0, None))

    # pylint: disable=broad-except
    def _publish_message(self, msg):
        _LOGGER_MSG.debug("RX: %s", repr(msg))
        topic = None
        kwargs = {}
        try:
            for (topic, kwargs) in convert_to_topic(msg):
                if _is_nak(msg) and not _has_listeners(topic):
                    self._resend(msg)
                else:
                    publish_topic(topic, **kwargs)
        except ValueError:
            # No topic was found for this message
            _LOGGER.debug("No topic found for message %r", msg)
        except Exception as ex:
            log_error(msg, ex, topic=topic, kwargs=kwargs)

    def _write(self, msg, priority=5):
        """Prepare data for writing to the transport.

        Data is actually writen by _write_message to ensure a pause beteen writes.
        This approach minimizes NAK messages. This also allows for some messages
        to be lower priority such as 'Load ALDB' versus higher priority such as
        'Set Light Level'.
        """
        self._message_queue.put_nowait((priority, msg))

    def _resend(self, msg):
        """Resend after a NAK message.

        TODO: Avoid resending the same message 10 times.
        """
        self._write(bytes(msg)[:-1])

    def _unsubscribe(self):
        """Unsubscribe to topics."""
        pub.unsubscribe(self._write, "send_message")

    def _start_writer(self):
        """Start the message writer."""
        asyncio.ensure_future(self._write_messages())

    async def _write_messages(self):
        """Write data to the transport."""
        _LOGGER.debug("Starting _write_messages")
        while not self._transport.is_closing():
            _LOGGER.debug("Started")
            _, msg = await self._message_queue.get()
            if msg:
                _LOGGER_MSG.debug("TX: %s", repr(msg))
                await self._transport.async_write(msg)
                await asyncio.sleep(WRITE_WAIT)
        _LOGGER.debug("Modem writer stopped.")



"""Create a topic and a direct message."""
import logging
from math import ceil

from .. import pub
from ..address import Address
from ..constants import RampRate
from ..utils import subscribe_topic
from ..topics import (
    ASSIGN_TO_ALL_LINK_GROUP,
    ASSIGN_TO_COMPANION_GROUP,
    BRIGHTEN_ONE_STEP,
    DELETE_FROM_ALL_LINK_GROUP,
    DEVICE_TEXT_STRING_REQUEST,
    DIM_ONE_STEP,
    DOOR_MOVE_CLOSE_DOOR,
    DOOR_MOVE_LOWER_DOOR,
    DOOR_MOVE_OPEN_DOOR,
    DOOR_MOVE_RAISE_DOOR,
    DOOR_MOVE_SINGLE_DOOR_CLOSE,
    DOOR_MOVE_SINGLE_DOOR_OPEN,
    DOOR_MOVE_STOP_DOOR,
    DOOR_STATUS_REPORT_CLOSE_DOOR,
    DOOR_STATUS_REPORT_LOWER_DOOR,
    DOOR_STATUS_REPORT_OPEN_DOOR,
    DOOR_STATUS_REPORT_RAISE_DOOR,
    DOOR_STATUS_REPORT_SINGLE_DOOR_CLOSE,
    DOOR_STATUS_REPORT_SINGLE_DOOR_OPEN,
    DOOR_STATUS_REPORT_STOP_DOOR,
    ENTER_LINKING_MODE,
    ENTER_UNLINKING_MODE,
    EXTENDED_GET_SET,
    EXTENDED_GET_SET_2,
    EXTENDED_READ_WRITE_ALDB,
    EXTENDED_TRIGGER_ALL_LINK,
    FX_USERNAME,
    GET_INSTEON_ENGINE_VERSION,
    GET_OPERATING_FLAGS,
    ID_REQUEST,
    INSTANT_CHANGE,
    IO_ALARM_DATA_REQUEST,
    IO_ALARM_DATA_RESPONSE,
    IO_GET_SENSOR_ALARM_DELTA,
    IO_GET_SENSOR_VALUE,
    IO_MODULE_DIAGNOSTICS_OFF,
    IO_MODULE_DIAGNOSTICS_ON,
    IO_MODULE_DISABLE_STATUS_CHANGE_MESSAGE,
    IO_MODULE_ENABLE_STATUS_CHANGE_MESSAGE,
    IO_MODULE_LOAD_EEPROM_FROM_RAM,
    IO_MODULE_LOAD_INITIALIZATION_VALUES,
    IO_MODULE_LOAD_RAM_FROM_EEPROM,
    IO_MODULE_READ_ANALOG_ALWAYS,
    IO_MODULE_READ_ANALOG_ONCE,
    IO_MODULE_SENSOR_OFF,
    IO_MODULE_SENSOR_ON,
    IO_MODULE_STATUS_REQUEST,
    IO_OUTPUT_OFF,
    IO_OUTPUT_ON,
    IO_READ_CONFIGURATION_PORT,
    IO_READ_INPUT_PORT,
    IO_SET_SENSOR_1_NOMINAL_VALUE,
    IO_SET_SENSOR_NOMINAL_VALUE,
    IO_WRITE_CONFIGURATION_PORT,
    IO_WRITE_OUTPUT_PORT,
    OFF,
    OFF_AT_RAMP_RATE,
    OFF_FAST,
    ON,
    ON_AT_RAMP_RATE,
    ON_FAST,
    PEEK_ONE_BYTE,
    PEEK_ONE_BYTE_INTERNAL,
    PING,
    POKE_ONE_BYTE,
    POKE_ONE_BYTE_INTERNAL,
    POOL_DEVICE_OFF,
    POOL_DEVICE_ON,
    POOL_GET_AMBIENT_TEMPERATURE,
    POOL_GET_PH,
    POOL_GET_POOL_MODE,
    POOL_GET_WATER_TEMPERATURE,
    POOL_LOAD_EEPROM_FROM_RAM,
    POOL_LOAD_INITIALIZATION_VALUES,
    POOL_SET_DEVICE_HYSTERESIS,
    POOL_SET_DEVICE_TEMPERATURE,
    POOL_TEMPERATURE_DOWN,
    POOL_TEMPERATURE_UP,
    PRODUCT_DATA_REQUEST,
    SET_ADDRESS_MSB,
    SET_ALL_LINK,
    SET_ALL_LINK_COMMAND_ALIAS,
    SET_DEVICE_TEXT_STRING,
    SET_OPERATING_FLAGS,
    SET_SPRINKLER_PROGRAM,
    SET_STATUS,
    SPRINKLER_BROADCAST_OFF,
    SPRINKLER_BROADCAST_ON,
    SPRINKLER_DIAGNOSTICS_OFF,
    SPRINKLER_DIAGNOSTICS_ON,
    SPRINKLER_DISABLE_PUMP_ON_V8,
    SPRINKLER_ENABLE_PUMP_ON_V8,
    SPRINKLER_GET_PROGRAM_REQUEST,
    SPRINKLER_GET_PROGRAM_RESPONSE,
    SPRINKLER_GET_VALVE_STATUS,
    SPRINKLER_INHIBIT_COMMAND_ACCEPTANCE,
    SPRINKLER_LOAD_EEPROM_FROM_RAM,
    SPRINKLER_LOAD_INITIALIZATION_VALUES,
    SPRINKLER_LOAD_RAM_FROM_EEPROM,
    SPRINKLER_PROGRAM_OFF,
    SPRINKLER_PROGRAM_ON,
    SPRINKLER_RESUME_COMMAND_ACCEPTANCE,
    SPRINKLER_SENSOR_OFF,
    SPRINKLER_SENSOR_ON,
    SPRINKLER_SKIP_BACK,
    SPRINKLER_SKIP_FORWARD,
    SPRINKLER_VALVE_OFF,
    SPRINKLER_VALVE_ON,
    STATUS_REQUEST,
    THERMOSTAT_GET_ZONE_INFORMATION,
    THERMOSTAT_CONTROL,
    THERMOSTAT_SET_COOL_SETPOINT,
    THERMOSTAT_SET_HEAT_SETPOINT,
    THERMOSTAT_SET_ZONE_COOL_SETPOINT,
    THERMOSTAT_SET_ZONE_HEAT_SETPOINT,
    THERMOSTAT_TEMPERATURE_DOWN,
    THERMOSTAT_TEMPERATURE_UP,
    THERMOSTAT_ZONE_TEMPERATURE_DOWN,
    THERMOSTAT_ZONE_TEMPERATURE_UP,
    WINDOW_COVERING_CLOSE,
    WINDOW_COVERING_OPEN,
    WINDOW_COVERING_POSITION,
    WINDOW_COVERING_PROGRAM,
    WINDOW_COVERING_STOP,
)
from .commands import commands
from .messages.all_link_record_flags import create
from .messages.message_flags import create as create_flags
from .messages.outbound import send_extended, send_standard
from .messages.user_data import UserData
from .topic_converters import topic_to_command_handler, topic_to_message_type

_LOGGER = logging.getLogger(__name__)
# pylint: disable=invalid-name
topic_register = {}


def register_command_handlers():
    """Register outbound handlers."""
    for topic in topic_register:
        func = topic_register[topic]
        subscribe_topic(func, topic)


# The following messages are all send_standard or send_extended messages
# The topis is based on the cmd1, cmd2 and extended message flags values


def _create_direct_message(
    topic, address, cmd2=None, user_data=None, crc=False, priority=5
):
    main_topic = topic.name.split(".")[1]
    command = commands.get_command(main_topic)
    extended = user_data is not None
    cmd2 = command.cmd2 if command.cmd2 is not None else cmd2
    flag_type = topic_to_message_type(topic)
    flags = create_flags(flag_type, extended)
    if extended:
        if crc:
            user_data.set_crc(command.cmd1, cmd2)
        else:
            user_data.set_checksum(command.cmd1, cmd2)
        send_extended(
            address=address,
            cmd1=command.cmd1,
            cmd2=cmd2,
            flags=flags,
            user_data=user_data,
            topic=topic,
            priority=priority,
        )
    else:
        send_standard(
            address=address,
            cmd1=command.cmd1,
            cmd2=cmd2,
            flags=flags,
            topic=topic,
            priority=priority,
        )


@topic_to_command_handler(register_list=topic_register, topic=ASSIGN_TO_ALL_LINK_GROUP)
def assign_to_all_link_group(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a ASSIGN_TO_ALL_LINK_GROUP command."""
    _create_direct_message(topic=topic, address=address, cmd2=group)


@topic_to_command_handler(
    register_list=topic_register, topic=DELETE_FROM_ALL_LINK_GROUP
)
def delete_from_all_link_group(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a DELETE_FROM_ALL_LINK_GROUP command."""
    _create_direct_message(topic=topic, address=address, cmd2=group)


@topic_to_command_handler(register_list=topic_register, topic=PRODUCT_DATA_REQUEST)
def product_data_request(address: Address, topic=pub.AUTO_TOPIC):
    """Create a PRODUCT_DATA_REQUEST command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=FX_USERNAME)
def fx_username(address: Address, topic=pub.AUTO_TOPIC):
    """Create a FX_USERNAME command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=DEVICE_TEXT_STRING_REQUEST
)
def device_text_string_request(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DEVICE_TEXT_STRING_REQUEST command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=SET_DEVICE_TEXT_STRING)
def set_device_text_string(address: Address, user_data, topic=pub.AUTO_TOPIC):
    """Create a SET_DEVICE_TEXT_STRING command."""
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(
    register_list=topic_register, topic=SET_ALL_LINK_COMMAND_ALIAS
)
def set_all_link_command_alias(address: Address, user_data, topic=pub.AUTO_TOPIC):
    """Create a SET_ALL_LINK_COMMAND_ALIAS command."""
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(register_list=topic_register, topic=SET_ALL_LINK)
def set_all_link(address: Address, user_data, topic=pub.AUTO_TOPIC):
    """Create a SET_ALL_LINK command."""
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(register_list=topic_register, topic=ENTER_LINKING_MODE)
def enter_linking_mode(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a ENTER_LINKING_MODE command."""
    user_data = UserData()
    _create_direct_message(
        topic=topic, address=address, cmd2=group, user_data=user_data
    )


@topic_to_command_handler(register_list=topic_register, topic=ENTER_UNLINKING_MODE)
def enter_unlinking_mode(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a ENTER_UNLINKING_MODE command."""
    _create_direct_message(topic=topic, address=address, cmd2=group)


@topic_to_command_handler(
    register_list=topic_register, topic=GET_INSTEON_ENGINE_VERSION
)
def get_insteon_engine_version(address: Address, topic=pub.AUTO_TOPIC):
    """Create a GET_INSTEON_ENGINE_VERSION command."""
    _create_direct_message(topic=topic, address=address, cmd2=0x00)


@topic_to_command_handler(register_list=topic_register, topic=PING)
def ping(address: Address, topic=pub.AUTO_TOPIC):
    """Create a PING command."""
    _create_direct_message(topic=topic, address=address, cmd2=0x00)


@topic_to_command_handler(register_list=topic_register, topic=ID_REQUEST)
def id_request(address: Address, topic=pub.AUTO_TOPIC):
    """Create a ID_REQUEST command."""
    _create_direct_message(topic=topic, address=address, cmd2=0x00)


@topic_to_command_handler(register_list=topic_register, topic=ON)
def on(address: Address, on_level: int, group=0, topic=pub.AUTO_TOPIC):
    """Create a ON command."""
    user_data = None
    if group and group > 1:
        user_data = UserData({"d1": group})
    _create_direct_message(
        topic=topic, address=address, cmd2=on_level, user_data=user_data
    )


@topic_to_command_handler(register_list=topic_register, topic=ON_FAST)
def on_fast(address: Address, on_level: int, group: int, topic=pub.AUTO_TOPIC):
    """Create a ON_FAST command."""
    user_data = None
    if group and group > 1:
        user_data = UserData({"d1": group})
    _create_direct_message(
        topic=topic, address=address, cmd2=on_level, user_data=user_data
    )


@topic_to_command_handler(register_list=topic_register, topic=OFF)
def off(address: Address, group: int, cmd2: int = 0, topic=pub.AUTO_TOPIC):
    """Create a OFF command."""
    user_data = None
    if group and group > 1:
        user_data = UserData({"d1": group})
    _create_direct_message(topic=topic, address=address, cmd2=cmd2, user_data=user_data)


@topic_to_command_handler(register_list=topic_register, topic=OFF_FAST)
def off_fast(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a OFF_FAST command."""
    user_data = None
    if group and group > 1:
        user_data = UserData({"d1": group})
    _create_direct_message(topic=topic, address=address, cmd2=0, user_data=user_data)


@topic_to_command_handler(register_list=topic_register, topic=BRIGHTEN_ONE_STEP)
def brighten_one_step(address: Address, topic=pub.AUTO_TOPIC):
    """Create a BRIGHTEN_ONE_STEP command."""
    _create_direct_message(topic=topic, address=address, cmd2=0)


@topic_to_command_handler(register_list=topic_register, topic=DIM_ONE_STEP)
def dim_one_step(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DIM_ONE_STEP command."""
    _create_direct_message(topic=topic, address=address, cmd2=0)


@topic_to_command_handler(register_list=topic_register, topic=STATUS_REQUEST)
def status_request(address: Address, status_type: int = 0, topic=pub.AUTO_TOPIC):
    """Create a STATUS_REQUEST command."""
    _create_direct_message(topic=topic, address=address, cmd2=status_type)


@topic_to_command_handler(register_list=topic_register, topic=GET_OPERATING_FLAGS)
def get_operating_flags(address: Address, flags_requested: int, topic=pub.AUTO_TOPIC):
    """Create a GET_OPERATING_FLAGS command."""
    _create_direct_message(topic=topic, address=address, cmd2=flags_requested)


@topic_to_command_handler(register_list=topic_register, topic=SET_OPERATING_FLAGS)
def set_operating_flags(
    address: Address, cmd: int, extended=False, topic=pub.AUTO_TOPIC
):
    """Create a SET_OPERATING_FLAGS command."""
    user_data = UserData() if extended else None
    _create_direct_message(topic=topic, address=address, cmd2=cmd, user_data=user_data)


@topic_to_command_handler(register_list=topic_register, topic=INSTANT_CHANGE)
def instant_change(address: Address, on_level: int, topic=pub.AUTO_TOPIC):
    """Create a INSTANT_CHANGE command."""
    _create_direct_message(topic=topic, address=address, cmd2=on_level)


@topic_to_command_handler(register_list=topic_register, topic=SET_STATUS)
def set_status(address: Address, on_level: int, topic=pub.AUTO_TOPIC):
    """Create a SET_STATUS command."""
    _create_direct_message(topic=topic, address=address, cmd2=on_level)


@topic_to_command_handler(register_list=topic_register, topic=SET_ADDRESS_MSB)
def set_address_msb(address: Address, high_byte: int, topic=pub.AUTO_TOPIC):
    """Create a SET_ADDRESS_MSB command."""
    _create_direct_message(topic=topic, address=address, cmd2=high_byte)


@topic_to_command_handler(register_list=topic_register, topic=POKE_ONE_BYTE)
def poke_one_byte(address: Address, byte_to_write: int, topic=pub.AUTO_TOPIC):
    """Create a POKE_ONE_BYTE command."""
    _create_direct_message(topic=topic, address=address, cmd2=byte_to_write)


@topic_to_command_handler(register_list=topic_register, topic=PEEK_ONE_BYTE)
def peek_one_byte(address: Address, lsb: int, topic=pub.AUTO_TOPIC):
    """Create a PEEK_ONE_BYTE command."""
    _create_direct_message(topic=topic, address=address, cmd2=lsb)


@topic_to_command_handler(register_list=topic_register, topic=PEEK_ONE_BYTE_INTERNAL)
def peek_one_byte_internal(address: Address, lsb: int, topic=pub.AUTO_TOPIC):
    """Create a PEEK_ONE_BYTE_INTERNAL command."""
    _create_direct_message(topic=topic, address=address, cmd2=lsb)


@topic_to_command_handler(register_list=topic_register, topic=POKE_ONE_BYTE_INTERNAL)
def poke_one_byte_internal(address: Address, byte_to_write: int, topic=pub.AUTO_TOPIC):
    """Create a POKE_ONE_BYTE_INTERNAL command."""
    _create_direct_message(topic=topic, address=address, cmd2=byte_to_write)


@topic_to_command_handler(register_list=topic_register, topic=ON_AT_RAMP_RATE)
def on_at_ramp_rate(
    address: Address, on_level: int, ramp_rate: RampRate, topic=pub.AUTO_TOPIC
):
    """Create a ON_AT_RAMP_RATE command."""
    on_level = min(0x10, on_level & 0xF0)
    ramp_rate = ceil(int(ramp_rate) / 2) + 1 & 0x0F
    cmd2 = on_level + ramp_rate
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(register_list=topic_register, topic=EXTENDED_GET_SET)
def extended_get_set(
    address: Address,
    data1=0,
    data2=0,
    data3=0,
    data4=0,
    data5=0,
    data6=0,
    data7=0,
    data8=0,
    data9=0,
    data10=0,
    data11=0,
    data12=0,
    data13=0,
    data14=0,
    crc=False,
    topic=pub.AUTO_TOPIC,
):
    """Create a EXTENDED_GET_SET command."""
    data = {}
    items = locals()
    for index in range(1, 15):
        data["d{}".format(index)] = items["data{}".format(index)]
    user_data = UserData(data)
    _create_direct_message(
        topic=topic, address=address, cmd2=0, user_data=user_data, crc=crc
    )


@topic_to_command_handler(register_list=topic_register, topic=EXTENDED_GET_SET_2)
def extended_get_set_2(
    address: Address,
    data1=0,
    data2=0,
    data3=0,
    data4=0,
    data5=0,
    data6=0,
    data7=0,
    data8=0,
    data9=0,
    data10=0,
    data11=0,
    data12=0,
    data13=0,
    data14=0,
    topic=pub.AUTO_TOPIC,
):
    """Create a EXTENDED_GET_SET_2 command."""
    data = {}
    items = locals()
    for index in range(1, 15):
        data["d{}".format(index)] = items["data{}".format(index)]
    user_data = UserData(data)
    _create_direct_message(
        topic=topic, address=address, cmd2=0x02, user_data=user_data, crc=True
    )


@topic_to_command_handler(register_list=topic_register, topic=OFF_AT_RAMP_RATE)
def off_at_ramp_rate(
    address: Address, on_level: int, ramp_rate: RampRate, topic=pub.AUTO_TOPIC
):
    """Create a OFF_AT_RAMP_RATE command."""
    on_level = min(0x10, on_level & 0xF0)
    ramp_rate = ceil(int(ramp_rate) / 2) + 1 & 0x0F
    cmd2 = on_level + ramp_rate
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


def _read_aldb(address, mem_addr, num_recs, topic):
    # num_recs = 0 if mem_addr == 0x0000 else 1
    mem_hi = mem_addr >> 8
    mem_lo = mem_addr & 0xFF
    user_data = UserData(
        {"d1": 0x00, "d2": 0x00, "d3": mem_hi, "d4": mem_lo, "d5": num_recs}
    )
    _create_direct_message(topic=topic, address=address, cmd2=0, user_data=user_data)


def _write_aldb(
    address,
    mem_addr,
    controller,
    group,
    target,
    data1,
    data2,
    data3,
    in_use,
    high_water_mark,
    bit5,
    bit4,
    topic,
):
    address = Address(address)
    target = Address(target)
    mem_hi = mem_addr >> 8
    mem_lo = mem_addr & 0xFF
    flags = create(
        in_use=in_use, controller=controller, hwm=high_water_mark, bit5=bit5, bit4=bit4
    )
    user_data = UserData(
        {
            "d2": 0x02,
            "d3": mem_hi,
            "d4": mem_lo,
            "d5": 0x08,
            "d6": int(flags),
            "d7": group,
            "d8": target.high,
            "d9": target.middle,
            "d10": target.low,
            "d11": data1,
            "d12": data2,
            "d13": data3,
        }
    )
    _create_direct_message(topic=topic, address=address, cmd2=0, user_data=user_data)


@topic_to_command_handler(register_list=topic_register, topic=EXTENDED_READ_WRITE_ALDB)
def extended_read_write_aldb(
    address: Address,
    action: int,
    mem_addr: int,
    num_recs: int = 0,
    controller: bool = True,
    group: int = 0x01,
    target: Address = None,
    data1: int = 0x00,
    data2: int = 0x00,
    data3: int = 0x00,
    in_use: bool = True,
    high_water_mark: bool = False,
    bit5: int = 0,
    bit4: int = 0,
    topic=pub.AUTO_TOPIC,
):
    """Create a EXTENDED_READ_WRITE_ALDB command."""
    if action == 0x00:
        _read_aldb(address=address, mem_addr=mem_addr, num_recs=num_recs, topic=topic)
    elif action == 0x02:
        _write_aldb(
            address=address,
            mem_addr=mem_addr,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
            topic=topic,
        )


@topic_to_command_handler(register_list=topic_register, topic=EXTENDED_TRIGGER_ALL_LINK)
def extended_trigger_all_link(address: Address, user_data, topic=pub.AUTO_TOPIC):
    """Create a EXTENDED_TRIGGER_ALL_LINK command."""
    _create_direct_message(topic=topic, address=address, cmd2=0, user_data=user_data)


@topic_to_command_handler(register_list=topic_register, topic=SET_SPRINKLER_PROGRAM)
def set_sprinkler_program(
    address: Address, program: int, user_data, topic=pub.AUTO_TOPIC
):
    """Create a SET_SPRINKLER_PROGRAM command."""
    _create_direct_message(
        topic=topic, address=address, cmd2=program, user_data=user_data
    )


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_VALVE_ON)
def sprinkler_valve_on(address: Address, valve: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_VALVE_ON command."""
    _create_direct_message(topic=topic, address=address, cmd2=valve)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_GET_PROGRAM_RESPONSE
)
def sprinkler_get_program_response(
    address: Address, program: int, user_data, topic=pub.AUTO_TOPIC
):
    """Create a SPRINKLER_GET_PROGRAM_RESPONSE command."""
    _create_direct_message(
        topic=topic, address=address, cmd2=program, user_data=user_data
    )


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_VALVE_OFF)
def sprinkler_valve_off(address: Address, valve: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_VALVE_OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=valve)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_PROGRAM_ON)
def sprinkler_program_on(address: Address, program: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_PROGRAM_ON command."""
    _create_direct_message(topic=topic, address=address, cmd2=program)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_PROGRAM_OFF)
def sprinkler_program_off(address: Address, program: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_PROGRAM_OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=program)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_LOAD_INITIALIZATION_VALUES
)
def sprinkler_load_initialization_values(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_LOAD_INITIALIZATION_VALUES command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_LOAD_EEPROM_FROM_RAM
)
def sprinkler_load_eeprom_from_ram(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_LOAD_EEPROM_FROM_RAM command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_GET_VALVE_STATUS
)
def sprinkler_get_valve_status(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_GET_VALVE_STATUS command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_INHIBIT_COMMAND_ACCEPTANCE
)
def sprinkler_inhibit_command_acceptance(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_INHIBIT_COMMAND_ACCEPTANCE command."""
    user_data = UserData({"d1": 0x03})
    _create_direct_message(topic=topic, address=address, cmd2=0x44, user_data=user_data)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_RESUME_COMMAND_ACCEPTANCE
)
def sprinkler_resume_command_acceptance(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_RESUME_COMMAND_ACCEPTANCE command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_SKIP_FORWARD)
def sprinkler_skip_forward(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_SKIP_FORWARD command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_SKIP_BACK)
def sprinkler_skip_back(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_SKIP_BACK command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_ENABLE_PUMP_ON_V8
)
def sprinkler_enable_pump_on_v8(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_ENABLE_PUMP_ON_V8 command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_DISABLE_PUMP_ON_V8
)
def sprinkler_disable_pump_on_v8(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_DISABLE_PUMP_ON_V8 command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_BROADCAST_ON)
def sprinkler_broadcast_on(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_BROADCAST_ON command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_BROADCAST_OFF)
def sprinkler_broadcast_off(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_BROADCAST_OFF command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_LOAD_RAM_FROM_EEPROM
)
def sprinkler_load_ram_from_eeprom(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_LOAD_RAM_FROM_EEPROM command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_SENSOR_ON)
def sprinkler_sensor_on(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_SENSOR_ON command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_SENSOR_OFF)
def sprinkler_sensor_off(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_SENSOR_OFF command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_DIAGNOSTICS_ON)
def sprinkler_diagnostics_on(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_DIAGNOSTICS_ON command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=SPRINKLER_DIAGNOSTICS_OFF)
def sprinkler_diagnostics_off(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_DIAGNOSTICS_OFF command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=SPRINKLER_GET_PROGRAM_REQUEST
)
def sprinkler_get_program_request(address: Address, program: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_GET_PROGRAM_REQUEST command."""
    _create_direct_message(topic=topic, address=address, cmd2=program)


@topic_to_command_handler(register_list=topic_register, topic=IO_OUTPUT_ON)
def io_output_on(address: Address, output_num: int, topic=pub.AUTO_TOPIC):
    """Create a IO_OUTPUT_ON command."""
    _create_direct_message(topic=topic, address=address, cmd2=output_num)


@topic_to_command_handler(register_list=topic_register, topic=IO_OUTPUT_OFF)
def io_output_off(address: Address, output_num: int, topic=pub.AUTO_TOPIC):
    """Create a IO_OUTPUT_OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=output_num)


@topic_to_command_handler(register_list=topic_register, topic=IO_ALARM_DATA_REQUEST)
def io_alarm_data_request(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_ALARM_DATA_REQUEST command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=IO_WRITE_OUTPUT_PORT)
def io_write_output_port(address: Address, value: int, topic=pub.AUTO_TOPIC):
    """Create a IO_WRITE_OUTPUT_PORT command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=IO_READ_INPUT_PORT)
def io_read_input_port(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_READ_INPUT_PORT command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=IO_GET_SENSOR_VALUE)
def io_get_sensor_value(address: Address, sensor: int, topic=pub.AUTO_TOPIC):
    """Create a IO_GET_SENSOR_VALUE command."""
    _create_direct_message(topic=topic, address=address, cmd2=sensor)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_SET_SENSOR_1_NOMINAL_VALUE
)
def io_set_sensor_1_nominal_value(address: Address, value: int, topic=pub.AUTO_TOPIC):
    """Create a IO_SET_SENSOR_1_NOMINAL_VALUE command."""
    _create_direct_message(topic=topic, address=address, cmd2=value)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_SET_SENSOR_NOMINAL_VALUE
)
def io_set_sensor_nominal_value(
    address: Address, value: int, user_data, topic=pub.AUTO_TOPIC
):
    """Create a IO_SET_SENSOR_NOMINAL_VALUE command."""
    _create_direct_message(
        topic=topic, address=address, cmd2=value, user_data=user_data
    )


@topic_to_command_handler(register_list=topic_register, topic=IO_GET_SENSOR_ALARM_DELTA)
def io_get_sensor_alarm_delta(
    address: Address, sensor: int, delta: int, direction: int, topic=pub.AUTO_TOPIC
):
    """Create a IO_GET_SENSOR_ALARM_DELTA command."""
    sensor = sensor & 0x0F
    delta = delta & 0x07 << 4
    direction = 8 if bool(direction) else 0
    cmd2 = sensor + delta + direction
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(register_list=topic_register, topic=IO_ALARM_DATA_RESPONSE)
def io_alarm_data_response(address: Address, user_data, topic=pub.AUTO_TOPIC):
    """Create a IO_ALARM_DATA_RESPONSE command."""
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_WRITE_CONFIGURATION_PORT
)
def io_write_configuration_port(
    address: Address,
    bits_0_1: bool,
    bit_2: bool,
    bit_3: bool,
    bit_4: bool,
    bit_5: bool,
    bit_6: bool,
    bit_7: bool,
    topic=pub.AUTO_TOPIC,
):
    """Create a IO_WRITE_CONFIGURATION_PORT command."""
    cmd2 = (
        bit_7
        << 7 + bit_6
        << 6 + bit_5
        << 5 + bit_4
        << 4 + bit_3
        << 3 + bit_2
        << 2 + bits_0_1
        << 0
    )
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_READ_CONFIGURATION_PORT
)
def io_read_configuration_port(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_READ_CONFIGURATION_PORT command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_MODULE_LOAD_INITIALIZATION_VALUES
)
def io_module_load_initialization_values(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_LOAD_INITIALIZATION_VALUES command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_MODULE_LOAD_EEPROM_FROM_RAM
)
def io_module_load_eeprom_from_ram(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_LOAD_EEPROM_FROM_RAM command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=IO_MODULE_STATUS_REQUEST)
def io_module_status_request(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_STATUS_REQUEST command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_MODULE_READ_ANALOG_ONCE
)
def io_module_read_analog_once(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_READ_ANALOG_ONCE command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_MODULE_READ_ANALOG_ALWAYS
)
def io_module_read_analog_always(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_READ_ANALOG_ALWAYS command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_MODULE_ENABLE_STATUS_CHANGE_MESSAGE
)
def io_module_enable_status_change_message(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_ENABLE_STATUS_CHANGE_MESSAGE command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_MODULE_DISABLE_STATUS_CHANGE_MESSAGE
)
def io_module_disable_status_change_message(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_DISABLE_STATUS_CHANGE_MESSAGE command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=IO_MODULE_LOAD_RAM_FROM_EEPROM
)
def io_module_load_ram_from_eeprom(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_LOAD_RAM_FROM_EEPROM command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=IO_MODULE_SENSOR_ON)
def io_module_sensor_on(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_SENSOR_ON command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=IO_MODULE_SENSOR_OFF)
def io_module_sensor_off(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_SENSOR_OFF command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=IO_MODULE_DIAGNOSTICS_ON)
def io_module_diagnostics_on(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_DIAGNOSTICS_ON command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=IO_MODULE_DIAGNOSTICS_OFF)
def io_module_diagnostics_off(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_MODULE_DIAGNOSTICS_OFF command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=POOL_DEVICE_ON)
def pool_device_on(address: Address, device_num: int, topic=pub.AUTO_TOPIC):
    """Create a POOL_DEVICE_ON command."""
    _create_direct_message(topic=topic, address=address, cmd2=device_num)


@topic_to_command_handler(
    register_list=topic_register, topic=POOL_SET_DEVICE_TEMPERATURE
)
def pool_set_device_temperature(
    address: Address, device_num: int, user_data, topic=pub.AUTO_TOPIC
):
    """Create a POOL_SET_DEVICE_TEMPERATURE command."""
    _create_direct_message(
        topic=topic, address=address, cmd2=device_num, user_data=user_data
    )


@topic_to_command_handler(register_list=topic_register, topic=POOL_DEVICE_OFF)
def pool_device_off(address: Address, device_num: int, topic=pub.AUTO_TOPIC):
    """Create a POOL_DEVICE_OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=device_num)


@topic_to_command_handler(
    register_list=topic_register, topic=POOL_SET_DEVICE_HYSTERESIS
)
def pool_set_device_hysteresis(
    address: Address, device_num: int, user_data, topic=pub.AUTO_TOPIC
):
    """Create a POOL_SET_DEVICE_HYSTERESIS command."""
    _create_direct_message(
        topic=topic, address=address, cmd2=device_num, user_data=user_data
    )


@topic_to_command_handler(register_list=topic_register, topic=POOL_TEMPERATURE_UP)
def pool_temperature_up(address: Address, degrees: int, topic=pub.AUTO_TOPIC):
    """Create a POOL_TEMPERATURE_UP command."""
    cmd2 = degrees * 2
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(register_list=topic_register, topic=POOL_TEMPERATURE_DOWN)
def pool_temperature_down(address: Address, degrees: int, topic=pub.AUTO_TOPIC):
    """Create a POOL_TEMPERATURE_DOWN command."""
    cmd2 = degrees * 2
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(
    register_list=topic_register, topic=POOL_LOAD_INITIALIZATION_VALUES
)
def pool_load_initialization_values(address: Address, topic=pub.AUTO_TOPIC):
    """Create a POOL_LOAD_INITIALIZATION_VALUES command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=POOL_LOAD_EEPROM_FROM_RAM)
def pool_load_eeprom_from_ram(address: Address, topic=pub.AUTO_TOPIC):
    """Create a POOL_LOAD_EEPROM_FROM_RAM command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=POOL_GET_POOL_MODE)
def pool_get_pool_mode(address: Address, topic=pub.AUTO_TOPIC):
    """Create a POOL_GET_POOL_MODE command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=POOL_GET_AMBIENT_TEMPERATURE
)
def pool_get_ambient_temperature(address: Address, topic=pub.AUTO_TOPIC):
    """Create a POOL_GET_AMBIENT_TEMPERATURE command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=POOL_GET_WATER_TEMPERATURE
)
def pool_get_water_temperature(address: Address, topic=pub.AUTO_TOPIC):
    """Create a POOL_GET_WATER_TEMPERATURE command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=POOL_GET_PH)
def pool_get_ph(address: Address, topic=pub.AUTO_TOPIC):
    """Create a POOL_GET_PH command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=DOOR_MOVE_RAISE_DOOR)
def door_move_raise_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_MOVE_RAISE_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=DOOR_MOVE_LOWER_DOOR)
def door_move_lower_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_MOVE_LOWER_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=DOOR_MOVE_OPEN_DOOR)
def door_move_open_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_MOVE_OPEN_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=DOOR_MOVE_CLOSE_DOOR)
def door_move_close_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_MOVE_CLOSE_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=DOOR_MOVE_STOP_DOOR)
def door_move_stop_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_MOVE_STOP_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=DOOR_MOVE_SINGLE_DOOR_OPEN
)
def door_move_single_door_open(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_MOVE_SINGLE_DOOR_OPEN command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=DOOR_MOVE_SINGLE_DOOR_CLOSE
)
def door_move_single_door_close(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_MOVE_SINGLE_DOOR_CLOSE command."""
    user_data = UserData({"d1": 0x06})
    _create_direct_message(topic=topic, address=address, cmd2=0x58, user_data=user_data)


@topic_to_command_handler(
    register_list=topic_register, topic=DOOR_STATUS_REPORT_RAISE_DOOR
)
def door_status_report_raise_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_STATUS_REPORT_RAISE_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=DOOR_STATUS_REPORT_LOWER_DOOR
)
def door_status_reportlower_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_STATUS_REPORT_LOWER_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=DOOR_STATUS_REPORT_OPEN_DOOR
)
def door_status_report_open_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_STATUS_REPORT_OPEN_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=DOOR_STATUS_REPORT_CLOSE_DOOR
)
def door_status_report_close_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_STATUS_REPORT_CLOSE_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=DOOR_STATUS_REPORT_STOP_DOOR
)
def door_status_report_stop_door(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_STATUS_REPORT_STOP_DOOR command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=DOOR_STATUS_REPORT_SINGLE_DOOR_OPEN
)
def door_status_report_single_door_open(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_STATUS_REPORT_SINGLE_DOOR_OPEN command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=topic_register, topic=DOOR_STATUS_REPORT_SINGLE_DOOR_CLOSE
)
def door_status_report_single_door_close(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DOOR_STATUS_REPORT_SINGLE_DOOR_CLOSE command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=WINDOW_COVERING_OPEN)
def window_covering_open(address: Address, topic=pub.AUTO_TOPIC):
    """Create a WINDOW_COVERING_OPEN command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=WINDOW_COVERING_CLOSE)
def window_covering_close(address: Address, topic=pub.AUTO_TOPIC):
    """Create a WINDOW_COVERING_CLOSE command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=WINDOW_COVERING_STOP)
def window_covering_stop(address: Address, topic=pub.AUTO_TOPIC):
    """Create a WINDOW_COVERING_STOP command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=WINDOW_COVERING_PROGRAM)
def window_covering_program(address: Address, topic=pub.AUTO_TOPIC):
    """Create a WINDOW_COVERING_PROGRAM command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=topic_register, topic=WINDOW_COVERING_POSITION)
def window_covering_position(address: Address, position: int, topic=pub.AUTO_TOPIC):
    """Create a WINDOW_COVERING_POSITION command."""
    _create_direct_message(topic=topic, address=address, cmd2=position)


@topic_to_command_handler(register_list=topic_register, topic=THERMOSTAT_TEMPERATURE_UP)
def thermostat_temperature_up(address: Address, degrees: int, topic=pub.AUTO_TOPIC):
    """Create a THERMOSTAT_TEMPERATURE_UP command."""
    cmd2 = int(round(degrees * 2, 0))
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(
    register_list=topic_register, topic=THERMOSTAT_ZONE_TEMPERATURE_UP
)
def thermostat_zone_temperature_up(
    address: Address, zone: int, user_data, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_ZONE_TEMPERATURE_UP command."""
    _create_direct_message(topic=topic, address=address, cmd2=zone, user_data=user_data)


@topic_to_command_handler(
    register_list=topic_register, topic=THERMOSTAT_TEMPERATURE_DOWN
)
def thermostat_temperature_down(address: Address, degrees: int, topic=pub.AUTO_TOPIC):
    """Create a THERMOSTAT_TEMPERATURE_DOWN command."""
    cmd2 = int(round(degrees * 2, 0))
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(
    register_list=topic_register, topic=THERMOSTAT_ZONE_TEMPERATURE_DOWN
)
def thermostat_zone_temperature_down(
    address: Address, zone: int, user_data, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_ZONE_TEMPERATURE_DOWN command."""
    _create_direct_message(topic=topic, address=address, cmd2=zone, user_data=user_data)


@topic_to_command_handler(
    register_list=topic_register, topic=THERMOSTAT_GET_ZONE_INFORMATION
)
def thermostat_get_zone_information(
    address: Address, zone: int, info: int, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_GET_ZONE_INFORMATION command.

    zone: (int) 0 to 31

    info: (int)
        0 = Temperature
        1 = Setpoint
        2 = Deadband
        3 = Humidity
    """
    zone = zone & 0x0F
    info = info & 0x03 << 5
    cmd2 = info + zone
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(register_list=topic_register, topic=THERMOSTAT_CONTROL)
def thermostat_control(address: Address, mode: int, topic=pub.AUTO_TOPIC):
    """Create a THERMOSTAT_CONTROL command."""
    user_data = UserData()
    _create_direct_message(
        topic=topic, address=address, cmd2=int(mode), user_data=user_data
    )


@topic_to_command_handler(
    register_list=topic_register, topic=THERMOSTAT_SET_COOL_SETPOINT
)
def thermostat_set_cool_setpoint(address: Address, degrees: int, topic=pub.AUTO_TOPIC):
    """Create a THERMOSTAT_SET_COOL_SETPOINT command."""
    cmd2 = int(degrees * 2)
    user_data = UserData()
    _create_direct_message(topic=topic, address=address, cmd2=cmd2, user_data=user_data)


@topic_to_command_handler(
    register_list=topic_register, topic=THERMOSTAT_SET_ZONE_COOL_SETPOINT
)
def thermostat_set_zone_cool_setpoint(
    address: Address, zone: int, user_data, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_SET_ZONE_COOL_SETPOINT command."""
    _create_direct_message(topic=topic, address=address, cmd2=zone, user_data=user_data)


@topic_to_command_handler(
    register_list=topic_register, topic=THERMOSTAT_SET_HEAT_SETPOINT
)
def thermostat_set_heat_setpoint(address: Address, degrees: int, topic=pub.AUTO_TOPIC):
    """Create a THERMOSTAT_SET_HEAT_SETPOINT command."""
    cmd2 = int(degrees * 2)
    user_data = UserData()
    _create_direct_message(topic=topic, address=address, cmd2=cmd2, user_data=user_data)


@topic_to_command_handler(
    register_list=topic_register, topic=THERMOSTAT_SET_ZONE_HEAT_SETPOINT
)
def thermostat_set_zone_heat_setpoint(
    address: Address, zone: int, user_data, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_SET_ZONE_HEAT_SETPOINT command."""
    _create_direct_message(topic=topic, address=address, cmd2=zone, user_data=user_data)


@topic_to_command_handler(register_list=topic_register, topic=ASSIGN_TO_COMPANION_GROUP)
def assign_to_companion_group(address: Address, topic=pub.AUTO_TOPIC):
    """Create a ASSIGN_TO_COMPANION_GROUP command."""
    _create_direct_message(topic=topic, address=address)



"""Thermostat temperature up command."""
from .. import direct_ack_handler
from ...topics import THERMOSTAT_TEMPERATURE_DOWN
from .direct_command import DirectCommandHandlerBase


class TemperatureDownCommand(DirectCommandHandlerBase):
    """Manage an outbound THERMOSTAT_TEMPERATURE_DOWN command to a device."""

    def __init__(self, address):
        """Init the TemperatureUpCommand class."""
        super().__init__(topic=THERMOSTAT_TEMPERATURE_DOWN, address=address)

    # pylint: disable=arguments-differ
    def send(self, degrees=1):
        """Send the OFF command."""
        super().send(degrees=degrees)

    # pylint: disable=arguments-differ
    async def async_send(self, degrees=1):
        """Send the OFF command async."""
        return await super().async_send(degrees=degrees)

    @direct_ack_handler
    def handle_direct_ack(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the OFF response direct ACK."""
        self._call_subscribers(degrees=cmd2 * 0.5)



"""Base class for the All-Link Database."""
import logging
from abc import ABC, abstractmethod

from ..utils import publish_topic, subscribe_topic
from ..address import Address
from ..constants import ALDBStatus, ALDBVersion
from ..topics import (
    ALDB_VERSION,
    DEVICE_LINK_CONTROLLER_CREATED,
    DEVICE_LINK_CONTROLLER_REMOVED,
    DEVICE_LINK_RESPONDER_CREATED,
    DEVICE_LINK_RESPONDER_REMOVED,
)
from .aldb_record import ALDBRecord

_LOGGER = logging.getLogger(__name__)


class ALDBBase(ABC):
    """Represents a base class for a device All-Link database."""

    def __init__(self, address, version=ALDBVersion.V2, mem_addr=0x0FFF):
        """Instantiate the ALL-Link Database object."""
        self._records = {}
        self._status = ALDBStatus.EMPTY
        self._version = version

        self._address = Address(address)
        self._mem_addr = mem_addr
        self._cb_aldb_loaded = None
        self._read_manager = None
        self._dirty_records = []
        self._hwm_record = None
        subscribe_topic(self.update_version, f"{repr(self._address)}.{ALDB_VERSION}")

    def __len__(self):
        """Return the number of devices in the ALDB."""
        return len(self._records)

    def __iter__(self):
        """Iterate through each ALDB device record."""
        for key in sorted(self._records, reverse=True):
            yield key

    def __getitem__(self, mem_addr):
        """Fetch a device from the ALDB."""
        return self._records.get(mem_addr)

    @abstractmethod
    def __setitem__(self, mem_addr, record):
        """Add or Update a device in the ALDB."""

    def __repr__(self):
        """Human representation of a device from the ALDB."""
        attrs = vars(self)
        return ", ".join("%s: %r" % item for item in attrs.items())

    @property
    def address(self) -> Address:
        """Returnt the status of the device."""
        return self._address

    @property
    def status(self) -> ALDBStatus:
        """Return the ALDB load status."""
        return self._status

    @property
    def version(self) -> ALDBVersion:
        """Return the ALDB version."""
        return self._version

    @property
    def first_mem_addr(self):
        """Return the expected memory address of the first record."""
        return self._mem_addr

    @property
    def high_water_mark_mem_addr(self):
        """Return the High Water Mark record memory address."""
        if self._hwm_record:
            return self._hwm_record.mem_addr
        return None

    @property
    def is_loaded(self) -> bool:
        """Test if the ALDB is loaded."""
        if self._status == ALDBStatus.LOADING:
            loaded = self._calc_load_status()
            if loaded:
                self._status = ALDBStatus.LOADED
        return self._status == ALDBStatus.LOADED

    def clear(self):
        """Clear all records from the ALDB.

        This does not write to the device.
        """
        for mem in self._records:
            rec = self._records[mem]
            self._notify_change(rec, force_delete=True)
        self._records = {}
        self._dirty_records = []

    def get(self, mem_addr, default=None):
        """Get the record at address 'mem_addr'."""
        return self._records.get(mem_addr, default)

    def get_responders(self, group):
        """Return all responders to this device for a group."""
        for mem_addr in self._records:
            rec = self._records[mem_addr]
            if rec.is_controller and rec.group == group:
                yield rec.target

    def update_version(self, version):
        """Update the ALDB version number."""
        self._version = version

    @abstractmethod
    async def async_load(self, *args, **kwargs):
        """Load the All-Link Database."""

    def load_saved_records(self, status: ALDBStatus, records: [ALDBRecord]):
        """Load All-Link records from a dictionary of saved records."""
        if isinstance(status, ALDBStatus):
            self._status = status
        else:
            self._status = ALDBStatus(status)
        self.clear()
        for mem_addr in records:
            record = records[mem_addr]
            self._records[mem_addr] = record
            self._notify_change(record)
        if self.is_loaded and self._records:
            keys = list(self._records.keys())
            keys.sort(reverse=True)
            self._mem_addr = keys[0]

    def _notify_change(self, record, force_delete=False):
        target = record.target
        group = record.group
        is_in_use = True if force_delete else record.is_in_use
        if record.is_controller and is_in_use:
            topic = DEVICE_LINK_CONTROLLER_CREATED
        elif record.is_controller and not is_in_use:
            topic = DEVICE_LINK_CONTROLLER_REMOVED
        elif not record.is_controller and is_in_use:
            topic = DEVICE_LINK_RESPONDER_CREATED
        else:
            topic = DEVICE_LINK_RESPONDER_REMOVED

        self._send_change(topic, self._address, target, group)

    @classmethod
    def _send_change(cls, topic, controller, responder, group):
        publish_topic(topic, controller=controller, responder=responder, group=group)

    # pylint: disable=no-self-use
    def _calc_load_status(self):
        """Calculate the load status."""
        return False



#!/usr/bin/env python3
"""Setup for oyinsteon module."""
from setuptools import find_packages, setup


def readme():
    """Return README file as a string."""
    with open("DESCRIPTION.rst", "r") as readme_file:
        return readme_file.read()


setup(
    name="pyinsteon",
    version="1.0.0",
    author="The pyinsteon Development Team",
    author_email="pyinsteon@harrisnj.net",
    url="https://github.com/pyinsteon/pyinsteon",
    license="MIT License",
    packages=find_packages(exclude=["tests", "tests.*", "topics", "docs", "samples"]),
    scripts=[],
    description="Python API for controlling Insteon devices",
    long_description=readme(),
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        "pyserial",
        "pyserial-asyncio",
        "aiohttp",
        "pypubsub",
        "aiofiles",
        "async_generator",
        "pyyaml",
    ],
    entry_points={"console_scripts": ["insteon_tools = pyinsteon.tools:tools"]},
)



"""Switched Lighting Control devices (CATEGORY 0x02)."""
from functools import partial
from typing import Iterable

from ..constants import ResponseStatus
from ..events import OFF_EVENT, OFF_FAST_EVENT, ON_EVENT, ON_FAST_EVENT
from ..extended_property import (
    LED_DIMMING,
    X10_HOUSE,
    X10_UNIT,
    ON_MASK,
    OFF_MASK,
    NON_TOGGLE_MASK,
    NON_TOGGLE_ON_OFF_MASK,
    TRIGGER_GROUP_MASK,
)
from ..groups import (
    ON_OFF_OUTLET_BOTTOM,
    ON_OFF_OUTLET_TOP,
    ON_OFF_SWITCH,
    ON_OFF_SWITCH_A,
    ON_OFF_SWITCH_B,
    ON_OFF_SWITCH_C,
    ON_OFF_SWITCH_D,
    ON_OFF_SWITCH_E,
    ON_OFF_SWITCH_F,
    ON_OFF_SWITCH_G,
    ON_OFF_SWITCH_H,
    ON_OFF_SWITCH_MAIN,
)
from ..groups.on_off import OnOff
from ..handlers.to_device.set_leds import SetLedsCommandHandler
from ..handlers.to_device.status_request import StatusRequestCommand
from ..operating_flag import (
    DUAL_LINE_ON,
    KEY_BEEP_ON,
    LED_BLINK_ON_ERROR_ON,
    LED_BLINK_ON_TX_ON,
    LED_ON,
    MOMENTARY_LINE_ON,
    PROGRAM_LOCK_ON,
    RESUME_DIM_ON,
    REVERSED_ON,
    THREE_WAY_ON,
    RF_DISABLE_ON,
    POWERLINE_DISABLE_ON,
)
from .commands import SET_LEDS_COMMAND, STATUS_COMMAND, GET_LEDS_COMMAND
from .on_off_responder_base import OnOffResponderBase
from ..utils import bit_is_set, set_bit, multiple_status
from .on_off_controller_base import ON_LEVEL_MANAGER


class SwitchedLightingControl(OnOffResponderBase):
    """Switched Lighting Control device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
        state_name=ON_OFF_SWITCH,
        on_event_name=ON_EVENT,
        off_event_name=OFF_EVENT,
        on_fast_event_name=ON_FAST_EVENT,
        off_fast_event_name=OFF_FAST_EVENT,
    ):
        """Init the OnOffResponderBase class."""
        buttons = {1: ON_OFF_SWITCH} if buttons is None else buttons
        super().__init__(
            address,
            cat,
            subcat,
            firmware,
            description,
            model,
            buttons,
            on_event_name,
            off_event_name,
            on_fast_event_name,
            off_fast_event_name,
        )


class SwitchedLightingControl_ApplianceLinc(SwitchedLightingControl):
    """ApplianceLinc based dimmable lights."""

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9, is_reversed=True)


class SwitchedLightingControl_SwitchLinc(SwitchedLightingControl):
    """SwichLinc based dimmable lights."""

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9, is_reversed=True)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)
        self._add_operating_flag(LED_BLINK_ON_ERROR_ON, 5, 2, 0x14, 0x15)

        self._add_property(LED_DIMMING, 3, 3)
        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)


class SwitchedLightingControl_ToggleLinc(SwitchedLightingControl):
    """ToggleLinc based on/off lights."""

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)
        self._add_operating_flag(LED_BLINK_ON_ERROR_ON, 5, 2, 0x14, 0x15)

        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)


class SwitchedLightingControl_InLineLinc(SwitchedLightingControl_SwitchLinc):
    """InLineLinc based dimmable lights."""


class SwitchedLightingControl_OutletLinc(SwitchedLightingControl):
    """OutletLinc based dimmable lights."""

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9, is_reversed=True)

        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)


class SwitchedLightingControl_Micro(SwitchedLightingControl):
    """Micro switch based dimmable lights."""

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9, is_reversed=True)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)

        self._add_operating_flag(LED_BLINK_ON_ERROR_ON, 2, 2, 0x15, 0x14)

        self._add_operating_flag(DUAL_LINE_ON, 3, 0, 0x1E, 0x1F)
        self._add_operating_flag(MOMENTARY_LINE_ON, 3, 1, 0x20, 0x21)
        self._add_operating_flag(THREE_WAY_ON, 3, 2, 0x23, 0x22)
        self._add_operating_flag(REVERSED_ON, 3, 3, 0x25, 0x24)


class SwitchedLightingControl_DinRail(SwitchedLightingControl):
    """DINRail based dimmable lights."""

    def _register_operating_flags(self):
        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9, is_reversed=True)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)

        self._add_property(LED_DIMMING, 3, 3)
        self._add_property(X10_HOUSE, 5, None)
        self._add_property(X10_UNIT, 6, None)


class SwitchedLightingControl_KeypadLinc(SwitchedLightingControl):
    """KeypadLinc base class."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
    ):
        """Init the SwitchedLightingControl_KeypadLinc class."""
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )

    async def async_on(self, group: int = 0):
        """Turn on the button LED."""
        if group in [0, 1]:
            result = await super().async_on(group=group)
        else:
            kwargs = self._change_led_status(led=group, is_on=True)
            result = await self._handlers[SET_LEDS_COMMAND].async_send(**kwargs)
        if result == ResponseStatus.SUCCESS:
            self._update_leds(group=group, value=0xFF, event=ON_EVENT)
        elif result == ResponseStatus.UNCLEAR:
            await self.async_status()
        return result

    async def async_off(self, group: int = 0):
        """Turn off the button LED."""
        if group in [0, 1]:
            result = await super().async_off(group=group)
        else:
            kwargs = self._change_led_status(led=group, is_on=False)
            result = await self._handlers[SET_LEDS_COMMAND].async_send(**kwargs)
        if result == ResponseStatus.SUCCESS:
            self._update_leds(group=group, value=0, event=OFF_EVENT)
        elif result == ResponseStatus.UNCLEAR:
            await self.async_status()
        return result

    async def async_status(self, group=None):
        """Check the status of the device."""
        retries = 5
        status = ResponseStatus.UNSENT
        while retries and status != ResponseStatus.SUCCESS:
            status0 = await super().async_status()
            status1 = await self._handlers[GET_LEDS_COMMAND].async_send()
            status = multiple_status(status0, status1)
            retries -= 1
        return status

    async def async_set_radio_buttons(self, buttons: Iterable):
        """Set a group of buttons to act as radio buttons.

        This takes in a iterable set of buttons (eg. (3,4,5,6)) to act as radio buttons where
        no two buttons are on at the same time.
        """
        if len(buttons) < 2:
            raise IndexError("At least two buttons required.")
        await self.async_read_ext_properties()
        for button in buttons:
            if button not in self._buttons.keys():
                raise ValueError(f"Button {button} not in button list.")
            button_str = f"_{button}" if button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            on_mask_new_value = 0
            off_mask_new_value = 0
            for bit in range(0, 8):
                if bit + 1 in buttons:
                    on_mask_value = bit != button - 1
                    off_mask_value = bit != button - 1
                else:
                    on_mask_value = bit_is_set(on_mask.value, bit)
                    off_mask_value = bit_is_set(off_mask.value, bit)
                on_mask_new_value = set_bit(on_mask_new_value, bit, on_mask_value)
                off_mask_new_value = set_bit(off_mask_new_value, bit, off_mask_value)
            on_mask.new_value = on_mask_new_value
            off_mask.new_value = off_mask_new_value
        await self.async_write_ext_properties()

    async def async_set_toggle_mode(self, button: int, mode: int):
        """Set the toggle mode of a button.

        Usage:
            button: Integer of the button number
            mode: Integer of the mode
                0: Toggle
                1: Non-Toggle ON only
                2: Non-Toggle OFF only
        """
        if button not in self._buttons.keys():
            raise ValueError(f"Button {button} not in button list.")
        if mode not in [0, 1, 2]:
            raise ValueError(f"Mode {mode} invalid. Valid mode are [0, 1, 2]")
        await self.async_read_ext_properties(group=1)
        toggle_mask = self.properties[NON_TOGGLE_MASK]
        on_off_mask = self.properties[NON_TOGGLE_ON_OFF_MASK]
        if mode == 0:
            toggle_mask.new_value = set_bit(toggle_mask.value, button - 1, False)
            on_off_mask.new_value = set_bit(on_off_mask.value, button - 1, False)
        elif mode == 1:
            toggle_mask.new_value = set_bit(toggle_mask.value, button - 1, True)
            on_off_mask.new_value = set_bit(on_off_mask.value, button - 1, True)
        else:
            toggle_mask.new_value = set_bit(toggle_mask.value, button - 1, True)
            on_off_mask.new_value = set_bit(on_off_mask.value, button - 1, False)
        await self.async_write_ext_properties()

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[SET_LEDS_COMMAND] = SetLedsCommandHandler(address=self.address)
        self._handlers[GET_LEDS_COMMAND] = StatusRequestCommand(
            self._address, status_type=1
        )

    def _register_groups(self):
        super()._register_groups()
        for button in self._buttons:
            name = self._buttons[button]
            self._groups[button] = OnOff(name=name, address=self._address, group=button)

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._handlers[GET_LEDS_COMMAND].subscribe(self._led_status)
        for group in self._buttons:
            if self._groups.get(group) is not None:
                led_method = partial(self._led_follow_check, group=group)
                self._managers[group][ON_LEVEL_MANAGER].subscribe(led_method)

    def _led_follow_check(self, group, on_level):
        """Check the other LEDs to confirm if they follow the effected LED."""
        for button in self._buttons:
            if button == group:
                continue
            button_str = f"_{button}" if button != 1 else ""
            on_mask = self._properties[f"{ON_MASK}{button_str}"]
            off_mask = self._properties[f"{OFF_MASK}{button_str}"]
            follow = bit_is_set(on_mask, group)
            set_off = bit_is_set(off_mask, group)
            if follow:
                if set_off:
                    self._groups[button].value = 0
                else:
                    self._groups[button].value = on_level

    def _change_led_status(self, led, is_on):
        leds = {}
        for curr_led in range(1, 9):
            var = "group{}".format(curr_led)
            curr_group = self._groups.get(curr_led)
            curr_val = bool(curr_group.value) if curr_group else False
            leds[var] = is_on if curr_led == led else curr_val
        return leds

    def _update_leds(self, group, value, event):
        """Check if the LED is toggle or not and set value."""
        if not self._groups.get(group):
            return
        if self._properties[NON_TOGGLE_MASK].value is not None:
            non_toogle = bit_is_set(self._properties[NON_TOGGLE_MASK].value, group)
        else:
            non_toogle = False
        if non_toogle:
            self._groups[group].value = 0
        else:
            self._groups[group].value = value
        self._events[group][event].trigger(value)

    def _led_status(self, db_version, status):
        """Set the on level of the LED from a status command."""
        for bit in range(2, 9):
            state = self._groups.get(bit)
            if state:
                state.value = bit_is_set(status, bit - 1)

    def _register_operating_flags(self):
        """Register operating flags."""
        super()._register_operating_flags()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LED_BLINK_ON_TX_ON, 0, 1, 2, 3)
        self._add_operating_flag(RESUME_DIM_ON, 0, 2, 4, 5)
        self._add_operating_flag(LED_ON, 0, 4, 8, 9, is_reversed=True)
        self._add_operating_flag(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B)
        self._add_operating_flag(RF_DISABLE_ON, 0, 6, 0x0C, 0x0D)
        self._add_operating_flag(POWERLINE_DISABLE_ON, 0, 7, 0x0E, 0x0F)
        self._add_operating_flag(
            LED_BLINK_ON_ERROR_ON, 5, 2, 0x14, 0x15, is_reversed=True
        )

        self._add_property(LED_DIMMING, 9, 7, 1)
        self._add_property(NON_TOGGLE_MASK, 0x0A, 0x08)
        self._add_property(NON_TOGGLE_ON_OFF_MASK, 0x0D, 0x0B)
        self._add_property(TRIGGER_GROUP_MASK, 0x0E, 0x0C)
        for button in self._buttons:
            button_str = f"_{button}" if button != 1 else ""
            self._add_property(f"{ON_MASK}{button_str}", 3, 2, button)
            self._add_property(f"{OFF_MASK}{button_str}", 4, 3, button)
            self._add_property(f"{X10_HOUSE}{button_str}", 5, None, button)
            self._add_property(f"{X10_UNIT}{button_str}", 6, None, button)


class SwitchedLightingControl_KeypadLinc_6(SwitchedLightingControl_KeypadLinc):
    """KeypadLinc 6 button switch."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the SwitchedLightingControl_KeypadLinc_6 class."""
        buttons = {
            1: ON_OFF_SWITCH_MAIN,
            3: ON_OFF_SWITCH_A,
            4: ON_OFF_SWITCH_B,
            5: ON_OFF_SWITCH_C,
            6: ON_OFF_SWITCH_D,
        }
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            buttons=buttons,
        )


class SwitchedLightingControl_KeypadLinc_8(SwitchedLightingControl_KeypadLinc):
    """KeypadLinc 8 button switch."""

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the SwitchedLightingControl_KeypadLinc_8 class."""
        buttons = {
            1: ON_OFF_SWITCH_MAIN,
            2: ON_OFF_SWITCH_B,
            3: ON_OFF_SWITCH_C,
            4: ON_OFF_SWITCH_D,
            5: ON_OFF_SWITCH_E,
            6: ON_OFF_SWITCH_F,
            7: ON_OFF_SWITCH_G,
            8: ON_OFF_SWITCH_H,
        }
        super().__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            button=buttons,
        )


class SwitchedLightingControl_OnOffOutlet(SwitchedLightingControl_ApplianceLinc):
    """On/Off outlet model 2663-222 Switched Lighting Control.

    Device Class 0x02 subcat 0x39
    """

    TOP_GROUP = 1
    BOTTOM_GROUP = 2

    def __init__(self, address, cat, subcat, firmware=0x00, description="", model=""):
        """Init the SwitchedLightingControl_KeypadLinc class."""
        buttons = {1: ON_OFF_OUTLET_TOP, 2: ON_OFF_OUTLET_BOTTOM}
        super().__init__(
            address, cat, subcat, firmware, description, model, buttons=buttons
        )

    def status(self, group=None):
        """Request the status of the device."""
        if group is None:
            self._handlers[STATUS_COMMAND].send()
        if group in [1, 2]:
            self._handlers[group][STATUS_COMMAND].send()

    async def async_status(self, group=None):
        """Request the status of the device."""
        if group is None:
            return await self._handlers[STATUS_COMMAND].async_send()
        if group in [1, 2]:
            return await self._handlers[group][STATUS_COMMAND].async_send()

    def _register_handlers_and_managers(self):
        super()._register_handlers_and_managers()
        self._handlers[STATUS_COMMAND] = StatusRequestCommand(self._address, 1)

        if self._handlers.get(self.TOP_GROUP) is None:
            self._handlers[self.TOP_GROUP] = {}
        self._handlers[self.TOP_GROUP][STATUS_COMMAND] = StatusRequestCommand(
            self._address, 0
        )

        if self._handlers.get(self.BOTTOM_GROUP) is None:
            self._handlers[self.BOTTOM_GROUP] = {}
        self._handlers[self.BOTTOM_GROUP][STATUS_COMMAND] = StatusRequestCommand(
            self._address, 0
        )

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()

        self._handlers[self.TOP_GROUP][STATUS_COMMAND].subscribe(
            self._handle_top_status
        )
        self._handlers[self.TOP_GROUP][STATUS_COMMAND].subscribe(
            self._handle_bottom_status
        )

    def _handle_status(self, db_version, status):
        """Set the status of the top and bottom outlets state."""
        self._groups[self.TOP_GROUP].value = status & 0x01
        self._groups[self.BOTTOM_GROUP].value = status & 0x02

    def _handle_top_status(self, db_version, status):
        """Set the status of the top outlet."""
        self._groups[self.TOP_GROUP].value = status

    def _handle_bottom_status(self, db_version, status):
        """Set the status of the bottom outlet."""
        self._groups[self.BOTTOM_GROUP].value = status



"""Modem command to get next ALDB record."""
# pylint: disable=no-self-use
import logging

from . import ack_handler, nak_handler
from ..topics import GET_FIRST_ALL_LINK_RECORD
from .outbound_base import OutboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class GetFirstAllLinkRecordHandler(OutboundHandlerBase):
    """Handle Get First All Link Record commands."""

    def __init__(self):
        """Init the GetNextAldbRecordNak class."""
        super().__init__(topic=GET_FIRST_ALL_LINK_RECORD)
        _LOGGER.debug("Setup GetFirstAllLinkRecordHandler")

    @ack_handler()
    def handle_ack(self):
        """Handle the ACK message and return True."""

    @nak_handler
    def handle_nak(self, **kwargs):
        """Receive the NAK message and return False."""



"""Command line tools to interact with the Insteon devices."""

from .. import devices
from ..constants import ALDBStatus
from ..managers.scene_manager import async_add_device_to_scene
from ..utils import seconds_to_ramp_rate
from .tools_base import ToolsBase
from .advanced import AdvancedTools


class ToolsAldb(ToolsBase):
    """Command class to test interactivity."""

    async def do_load_aldb(self, *args, **kwargs):
        """Load the All-Link Database of a device.

        Usage:
            load_aldb <ADDRESS>|all y|n  Load one or all devices (can be the modem address)

        To clear the current ALDB and reload from the device, enter `y` as the second argment.
        Otherwise enter `n`.
        """
        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        try:
            refresh_yn = args[1]
            refresh = refresh_yn.lower() == "y"
        except IndexError:
            refresh_yn = ""

        addresses = await self._get_addresses(
            address=address, allow_cancel=True, allow_all=True,
        )
        if not addresses:
            return

        if devices[addresses[0]] != devices.modem or len(addresses) > 1:
            if not refresh_yn:
                refresh_yn = await self._get_char(
                    "Clear existing records and reload (y/n)",
                    default="n",
                    values=["y", "n"],
                )
        refresh = refresh_yn.lower() == "y"

        battery_devices = []
        for address in addresses:
            if devices[address].is_battery:
                battery_devices.append(address)
            # Only load the modem ALDB if explicitly asked
            if devices[address] == devices.modem and len(addresses) == 1:
                await devices.modem.aldb.async_load()
            elif devices[address].cat == 0x03:
                pass
            else:
                # tasks.append(devices[address].aldb.async_load(refresh=refresh))
                await devices[address].aldb.async_load(refresh=refresh)

        if battery_devices:
            self._log_stdout("The following devices are battery operated.")
            self._log_stdout("They will load in the background when they wake up.")
        for address in battery_devices:
            self._log_stdout(f"    - {address}")

        # if the device did not load the first time, try one more time with refresh
        for address in addresses:
            if (
                devices[address] != devices.modem
                and devices[address].aldb.status != ALDBStatus.LOADED
                and not devices[address].is_battery
            ):
                await devices[address].aldb.async_load(refresh=refresh)

    async def do_print_aldb(self, *args, **kwargs):
        """Print the records in an All-Link Database.

        Usage:
            print_aldb <ADDRESS>|all
        """
        await self._print_aldb(*args)

    async def do_add_default_links(self, *args, **kwargs):
        """Add default links between a device and the modem.

        Usage:
            add_default_links <ADDRESS>
        """

        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        addresses = await self._get_addresses(
            address=address, allow_all=False, allow_cancel=True,
        )
        if not addresses:
            return
        device = devices[addresses[0]]
        self._log_command(f"add_default_links {addresses[0]}")
        await device.async_add_default_links()

    async def do_add_device_to_scene(self, *args, **kwargs):
        """Add a device to a scene.

        Usage:
            add_device_to_scene <ADDRESS> <SCENE NUMBER> [<ON LEVEL>] [<RAMP RATE>] [<BUTTON>]
        <ADDRESS>: The device address such as 1a.2b.3c
        <SCENE NUMBER>: Value from 25 to 255.
        <ON LEVEL>: (Optional) Value from 0 (off) - 255 (full on).
                    For dimmable devices any number from 0 to 255 is allowable.
                    For on/off devices only 0 and 255 are allowed.
                    Default is 255.
        <RAMP RATE>: (Only valid for dimmable devices) 0.1 seconds to 480 seconds (8 minutes)
                    Default is 0.5 seconds
        <BUTTON>: (Optional) The button or group number of the device to change as part of the scene.
                    Valid values are device dependant.
                    Default is 1.

        KeyPadLinc devices will not respond correctly to scenes in this way other than the main power.
        """

        args = args[0].split()
        try:
            address = args[0]
        except IndexError:
            address = None

        try:
            scene = int(args[1])
        except (IndexError, ValueError):
            scene = None

        try:
            data1 = int(args[2])
        except (IndexError, ValueError):
            data1 = None

        try:
            data2_seconds = int(args[3])
        except (IndexError, ValueError):
            data2_seconds = None

        try:
            data3 = int(args[4])
        except (IndexError, ValueError):
            data3 = None

        addresses = await self._get_addresses(
            address=address, allow_all=False, allow_cancel=True,
        )
        if not addresses:
            return
        if not scene:
            scene = await self._get_int(
                "Scene number or blank to cancel", values=range(25, 256),
            )
            if not scene:
                return

        if not data1:
            data1 = await self._get_int("On level", default=255, values=range(25, 256),)

        if not data2_seconds:
            data2_seconds = await self._get_float(
                "Ramp rate", default=0.5, maximum=480, minimum=0.1,
            )

        if not data3:
            data2_seconds = await self._get_int("Button", default=1)

        data2 = seconds_to_ramp_rate(data2_seconds)
        device = devices[addresses[0]]
        await async_add_device_to_scene(device, scene, data1, data2, data3)

    def do_print_aldb_load_status(self, *args, **kwargs):
        """Print the All-Link databbase load status for all devices."""
        self._log_stdout("")
        self._log_stdout("Device   Status")
        self._log_stdout("-------- ---------------")
        for address in devices:
            self._log_stdout(f"{address} {str(devices[address].aldb.status)}")

    async def do_advanced(self, *args, **kwargs):
        """Enter advanced ALDB menu."""
        self._log_command("advanced")
        await self._call_next_menu(AdvancedTools, "advanced")
